from jinja2 import Environment, FileSystemLoader
import yaml
import os
import sys
import stringcase
import shutil
import uncrustify
import clang_format
from pathlib import Path
import generation_dict_tools as tools


def get_generation_dict(
    common_cmake_folder_path: str,
    application_name: str,
    feature_by_name: dict,
    entities_by_name: dict,
) -> dict:
    generation_dict = {}
    for feature_name, feature in feature_by_name.items():
        feature_snake_name = stringcase.snakecase(feature_name)
        feature_pascal_name = stringcase.pascalcase(feature_name)
        feature_spinal_name = stringcase.spinalcase(feature_name)
        feature_camel_name = stringcase.camelcase(feature_name)
        generation_dict[feature_pascal_name] = {
            "feature_snake_name": feature_snake_name,
            "feature_pascal_name": feature_pascal_name,
            "feature_spinal_name": feature_spinal_name,
            "feature_camel_name": feature_camel_name,
        }
        # add export_header
        export_header = f"application_{feature_snake_name}_export.h"
        generation_dict[feature_pascal_name]["export_header"] = export_header
        generation_dict[feature_pascal_name]["export_header_file"] = os.path.join(
            common_cmake_folder_path,
            feature_snake_name + "_feature",
            export_header,
        )
        generation_dict[feature_pascal_name][
            "export"
        ] = f"{stringcase.uppercase(application_name)}_APPLICATION_{stringcase.uppercase(feature_snake_name)}_EXPORT"

    # add CRUD handlers
    for feature_name, feature in generation_dict.items():
        feature_snake_name = feature["feature_snake_name"]
        feature_pascal_name = feature["feature_pascal_name"]
        feature["crud_handlers"] = {}
        crud_handlers = feature["crud_handlers"]
        crud_data = feature_by_name[feature_name].get("CRUD", {})
        entity_mappable_with = crud_data.get("entity_mappable_with", None)
        if crud_data.get("enabled", False) and not entity_mappable_with:
            raise Exception(
                f"CRUD.entity_mappable_with not defined for feature {feature_name}"
            )
        entity_mappable_with_snake = stringcase.snakecase(entity_mappable_with)
        entity_mappable_with_pascal = stringcase.pascalcase(entity_mappable_with)
        if crud_data.get("enabled", False):
            create_data = crud_data.get("create", {})
            if create_data.get("enabled", False) and create_data.get("generate", True):
                # find out if the owner field is a list

                owner_dict = tools.determine_owner(
                    entity_mappable_with_pascal, entities_by_name
                )

                # fill crud_handlers for create
                crud_handlers["create"] = {
                    "generate": True,
                    "templates": [
                        "create_handler.cpp.jinja2",
                        "create_handler.h.jinja2",
                    ],
                    "files": [
                        os.path.join(
                            common_cmake_folder_path,
                            feature_snake_name + "_feature",
                            feature_snake_name,
                            "commands",
                            f"create_{entity_mappable_with_snake}_command_handler.cpp",
                        ),
                        os.path.join(
                            common_cmake_folder_path,
                            feature_snake_name + "_feature",
                            feature_snake_name,
                            "commands",
                            f"create_{entity_mappable_with_snake}_command_handler.h",
                        ),
                    ],
                    "entity_mappable_with_snake": entity_mappable_with_snake,
                    "entity_mappable_with_pascal": entity_mappable_with_pascal,
                    "entity_mappable_with_camel": stringcase.camelcase(
                        entity_mappable_with
                    ),
                    # unique to create : need to add the new entity to the owner entity
                    "has_owner": True if owner_dict else False,
                    "owner_name_snake": stringcase.snakecase(
                        owner_dict.get("name", "")
                    ),
                    "owner_name_pascal": stringcase.pascalcase(
                        owner_dict.get("name", "")
                    ),
                    "owner_name_camel": stringcase.camelcase(
                        owner_dict.get("name", "")
                    ),
                    "owner_field_name_snake": stringcase.snakecase(
                        owner_dict.get("field", "")
                    ),
                    "owner_field_name_pascal": stringcase.pascalcase(
                        owner_dict.get("field", "")
                    ),
                    "owner_field_name_camel": stringcase.camelcase(
                        owner_dict.get("field", "")
                    ),
                    "owner_field_is_list": owner_dict.get("is_list", False),
                    "owner_field_is_ordered": owner_dict.get("ordered", False),
                }
            remove_data = crud_data.get("remove", {})
            if remove_data.get("enabled", False) and remove_data.get("generate", True):
                crud_handlers["remove"] = {
                    "generate": True,
                    "templates": [
                        "remove_handler.cpp.jinja2",
                        "remove_handler.h.jinja2",
                    ],
                    "files": [
                        os.path.join(
                            common_cmake_folder_path,
                            feature_snake_name + "_feature",
                            feature_snake_name,
                            "commands",
                            f"remove_{entity_mappable_with_snake}_command_handler.cpp",
                        ),
                        os.path.join(
                            common_cmake_folder_path,
                            feature_snake_name + "_feature",
                            feature_snake_name,
                            "commands",
                            f"remove_{entity_mappable_with_snake}_command_handler.h",
                        ),
                    ],
                    "entity_mappable_with_snake": entity_mappable_with_snake,
                    "entity_mappable_with_pascal": entity_mappable_with_pascal,
                    "entity_mappable_with_camel": stringcase.camelcase(
                        entity_mappable_with
                    ),
                }
            update_data = crud_data.get("update", {})
            if update_data.get("enabled", False) and update_data.get("generate", True):
                crud_handlers["update"] = {
                    "generate": True,
                    "templates": [
                        "update_handler.cpp.jinja2",
                        "update_handler.h.jinja2",
                    ],
                    "files": [
                        os.path.join(
                            common_cmake_folder_path,
                            feature_snake_name + "_feature",
                            feature_snake_name,
                            "commands",
                            f"update_{entity_mappable_with_snake}_command_handler.cpp",
                        ),
                        os.path.join(
                            common_cmake_folder_path,
                            feature_snake_name + "_feature",
                            feature_snake_name,
                            "commands",
                            f"update_{entity_mappable_with_snake}_command_handler.h",
                        ),
                    ],
                    "entity_mappable_with_snake": entity_mappable_with_snake,
                    "entity_mappable_with_pascal": entity_mappable_with_pascal,
                    "entity_mappable_with_camel": stringcase.camelcase(
                        entity_mappable_with
                    ),
                }
            get_data = crud_data.get("get", {})
            if get_data.get("enabled", False) and get_data.get("generate", True):
                crud_handlers["get"] = {
                    "generate": True,
                    "templates": [
                        "get_handler.cpp.jinja2",
                        "get_handler.h.jinja2",
                    ],
                    "files": [
                        os.path.join(
                            common_cmake_folder_path,
                            feature_snake_name + "_feature",
                            feature_snake_name,
                            "queries",
                            f"get_{entity_mappable_with_snake}_query_handler.cpp",
                        ),
                        os.path.join(
                            common_cmake_folder_path,
                            feature_snake_name + "_feature",
                            feature_snake_name,
                            "queries",
                            f"get_{entity_mappable_with_snake}_query_handler.h",
                        ),
                    ],
                    "entity_mappable_with_snake": entity_mappable_with_snake,
                    "entity_mappable_with_pascal": entity_mappable_with_pascal,
                    "entity_mappable_with_camel": stringcase.camelcase(
                        entity_mappable_with
                    ),
                }
            get_all_data = crud_data.get("get_all", {})
            if get_all_data.get("enabled", False) and get_all_data.get(
                "generate", True
            ):
                crud_handlers["get_all"] = {
                    "generate": True,
                    "templates": [
                        "get_all_handler.cpp.jinja2",
                        "get_all_handler.h.jinja2",
                    ],
                    "files": [
                        os.path.join(
                            common_cmake_folder_path,
                            feature_snake_name + "_feature",
                            feature_snake_name,
                            "queries",
                            f"get_all_{entity_mappable_with_snake}_query_handler.cpp",
                        ),
                        os.path.join(
                            common_cmake_folder_path,
                            feature_snake_name + "_feature",
                            feature_snake_name,
                            "queries",
                            f"get_all_{entity_mappable_with_snake}_query_handler.h",
                        ),
                    ],
                    "entity_mappable_with_snake": entity_mappable_with_snake,
                    "entity_mappable_with_pascal": entity_mappable_with_pascal,
                    "entity_mappable_with_camel": stringcase.camelcase(
                        entity_mappable_with
                    ),
                }
            get_with_details_data = crud_data.get("get_with_details", {})
            if get_with_details_data.get(
                "enabled", False
            ) and get_with_details_data.get("generate", True):
                crud_handlers["get_with_details"] = {
                    "generate": True,
                    "templates": [
                        "get_with_details_handler.cpp.jinja2",
                        "get_with_details_handler.h.jinja2",
                    ],
                    "files": [
                        os.path.join(
                            common_cmake_folder_path,
                            feature_snake_name + "_feature",
                            feature_snake_name,
                            "queries",
                            f"get_{entity_mappable_with_snake}_with_details_query_handler.cpp",
                        ),
                        os.path.join(
                            common_cmake_folder_path,
                            feature_snake_name + "_feature",
                            feature_snake_name,
                            "queries",
                            f"get_{entity_mappable_with_snake}_with_details_query_handler.h",
                        ),
                    ],
                    "lazy_load_pascal_fields": [
                        stringcase.pascalcase(field)
                        for field in get_lazy_loading_fields(
                            entity_mappable_with_pascal, entities_by_name
                        )
                    ],
                    "entity_mappable_with_snake": entity_mappable_with_snake,
                    "entity_mappable_with_pascal": entity_mappable_with_pascal,
                    "entity_mappable_with_camel": stringcase.camelcase(
                        entity_mappable_with
                    ),
                }
        for handler_name, handler_data in crud_handlers.items():
            handler_data["export"] = feature["export"]
            handler_data["export_header"] = feature["export_header"]
            handler_data["export_header_file"] = feature["export_header_file"]

    # add commands and queries to the feature:
    for feature_name, feature in generation_dict.items():
        feature_snake_name = feature["feature_snake_name"]
        feature_pascal_name = feature["feature_pascal_name"]
        feature["custom_handlers"] = {}
        if feature_by_name[feature_name].get("commands", None):
            feature["custom_handlers"]["commands"] = feature_by_name[feature_name][
                "commands"
            ]

            for command in feature["custom_handlers"]["commands"]:
                files = [
                    os.path.join(
                        common_cmake_folder_path,
                        feature_snake_name + "_feature",
                        feature_snake_name,
                        "commands",
                        f"{stringcase.snakecase(command['name'])}_command_handler.h",
                    ),
                    os.path.join(
                        common_cmake_folder_path,
                        feature_snake_name + "_feature",
                        feature_snake_name,
                        "commands",
                        f"{stringcase.snakecase(command['name'])}_command_handler.cpp",
                    ),
                ]
                command["files"] = files
                command["templates"] = [
                    "custom_command_handler.h.jinja2",
                    "custom_command_handler.cpp.jinja2",
                ]
                command["repositories"] = []
                for entity_name in command.get("entities", []):
                    command["repositories"].append(
                        {
                            "pascal_name": stringcase.pascalcase(entity_name),
                            "snake_name": stringcase.snakecase(entity_name),
                            "camel_name": stringcase.camelcase(entity_name),
                        }
                    )

                command["export"] = feature["export"]
                command["export_header"] = feature["export_header"]
                command["export_header_file"] = feature["export_header_file"]
                command["camel_name"] = stringcase.camelcase(command["name"])
                command["snake_name"] = stringcase.snakecase(command["name"])
                command["pascal_name"] = stringcase.pascalcase(command["name"])
                command["dto_in_is_enabled"] = True
                # DTO in
                if command.get("dto", {}).get("in", {}):
                    if command.get("dto", {}).get("in", {}).get("enabled", True):
                        command["dto_in_pascal_name"] = stringcase.pascalcase(
                            command["dto"]["in"]["type_prefix"] + "DTO"
                        )
                        command["dto_in_camel_name"] = stringcase.camelcase(
                            command["dto"]["in"]["type_prefix"] + "DTO"
                        )
                        command["dto_in_snake_name"] = (
                            stringcase.snakecase(command["dto"]["in"]["type_prefix"])
                            + "_dto"
                        )
                    else:
                        command["dto_in_pascal_name"] = "void"
                        command["dto_in_camel_name"] = "void"
                        command["dto_in_snake_name"] = "void"
                        command["dto_in_is_enabled"] = False
                # DTO out
                command["dto_out_is_enabled"] = True

                if command.get("dto", {}).get("out", {}):
                    if command.get("dto", {}).get("out", {}).get("enabled", True):
                        command["dto_out_pascal_name"] = stringcase.pascalcase(
                            command["dto"]["out"]["type_prefix"] + "DTO"
                        )
                        command["dto_out_camel_name"] = stringcase.camelcase(
                            command["dto"]["out"]["type_prefix"] + "DTO"
                        )
                        command["dto_out_snake_name"] = (
                            stringcase.snakecase(command["dto"]["out"]["type_prefix"])
                            + "_dto"
                        )
                    else:
                        command["dto_out_pascal_name"] = "void"
                        command["dto_out_camel_name"] = "void"
                        command["dto_out_snake_name"] = "void"
                        command["dto_out_is_enabled"] = False

                # for each command.dto.out.type_prefix, add command.dto.out.type_prefix_camel
                for _, dto_data in command.get("dto", {}).items():
                    if dto_data.get("type_prefix", None):
                        dto_data["type_prefix_camel"] = stringcase.camelcase(
                            dto_data["type_prefix"]
                        )
                        dto_data["type_prefix_snake"] = stringcase.snakecase(
                            dto_data["type_prefix"]
                        )
                        dto_data["type_prefix_pascal"] = stringcase.pascalcase(
                            dto_data["type_prefix"]
                        )

        if feature_by_name[feature_name].get("queries", None):
            feature["custom_handlers"]["queries"] = feature_by_name[feature_name][
                "queries"
            ]

            for query in feature["custom_handlers"]["queries"]:
                files = [
                    os.path.join(
                        common_cmake_folder_path,
                        feature_snake_name + "_feature",
                        feature_snake_name,
                        "queries",
                        f"{stringcase.snakecase(query['name'])}_query_handler.h",
                    ),
                    os.path.join(
                        common_cmake_folder_path,
                        feature_snake_name + "_feature",
                        feature_snake_name,
                        "queries",
                        f"{stringcase.snakecase(query['name'])}_query_handler.cpp",
                    ),
                ]
                query["files"] = files
                query["templates"] = [
                    "custom_query_handler.h.jinja2",
                    "custom_query_handler.cpp.jinja2",
                ]

                query["repositories"] = []
                for entity_name in query.get("entities", []):
                    query["repositories"].append(
                        {
                            "pascal_name": stringcase.pascalcase(entity_name),
                            "snake_name": stringcase.snakecase(entity_name),
                            "camel_name": stringcase.camelcase(entity_name),
                        }
                    )
                query["export"] = feature["export"]
                query["export_header"] = feature["export_header"]
                query["export_header_file"] = feature["export_header_file"]
                query["camel_name"] = stringcase.camelcase(query["name"])
                query["snake_name"] = stringcase.snakecase(query["name"])
                query["pascal_name"] = stringcase.pascalcase(query["name"])
                query["dto_in_is_enabled"] = True
                # DTO in
                if query.get("dto", {}).get("in", {}):
                    if query.get("dto", {}).get("in", {}).get("enabled", True):
                        query["dto_in_pascal_name"] = stringcase.pascalcase(
                            query["dto"]["in"]["type_prefix"] + "DTO"
                        )
                        query["dto_in_camel_name"] = stringcase.camelcase(
                            query["dto"]["in"]["type_prefix"] + "DTO"
                        )
                        query["dto_in_snake_name"] = (
                            stringcase.snakecase(query["dto"]["in"]["type_prefix"])
                            + "_dto"
                        )
                    else:
                        query["dto_in_pascal_name"] = "void"
                        query["dto_in_camel_name"] = "void"
                        query["dto_in_snake_name"] = "void"
                        query["dto_in_is_enabled"] = False
                # DTO out
                query["dto_out_is_enabled"] = True

                if query.get("dto", {}).get("out", {}):
                    if query.get("dto", {}).get("out", {}).get("enabled", True):
                        query["dto_out_pascal_name"] = stringcase.pascalcase(
                            query["dto"]["out"]["type_prefix"] + "DTO"
                        )
                        query["dto_out_camel_name"] = stringcase.camelcase(
                            query["dto"]["out"]["type_prefix"] + "DTO"
                        )
                        query["dto_out_snake_name"] = (
                            stringcase.snakecase(query["dto"]["out"]["type_prefix"])
                            + "_dto"
                        )
                    else:
                        query["dto_out_pascal_name"] = "void"
                        query["dto_out_camel_name"] = "void"
                        query["dto_out_snake_name"] = "void"
                        query["dto_out_is_enabled"] = False

                # for each query.dto.out.type_prefix, add query.dto.out.type_prefix_camel
                for _, dto_data in query.get("dto", {}).items():
                    if dto_data.get("type_prefix", None):
                        dto_data["type_prefix_camel"] = stringcase.camelcase(
                            dto_data["type_prefix"]
                        )
                        dto_data["type_prefix_snake"] = stringcase.snakecase(
                            dto_data["type_prefix"]
                        )
                        dto_data["type_prefix_pascal"] = stringcase.pascalcase(
                            dto_data["type_prefix"]
                        )

        # add commands and queries files to feature custom handlers:
        feature["custom_handlers"]["files"] = []
        if feature["custom_handlers"].get("commands", None):
            for command in feature["custom_handlers"]["commands"]:
                feature["custom_handlers"]["files"] += command["files"]
        if feature["custom_handlers"].get("queries", None):
            for query in feature["custom_handlers"]["queries"]:
                feature["custom_handlers"]["files"] += query["files"]

    # add files to the feature:
    for feature_name, feature in generation_dict.items():
        feature_snake_name = stringcase.snakecase(feature_name)
        feature["handler_files"] = []
        for crud_handler_name, crud_handler in feature["crud_handlers"].items():
            feature["handler_files"] += crud_handler["files"]

        # add custom handlers files to the feature:
        if feature.get("custom_handlers", None):
            custom_handler = feature["custom_handlers"]
            feature["handler_files"] += custom_handler["files"]

        # add CMakelists.txt file name to the feature:
        feature["cmakelists_file"] = os.path.join(
            common_cmake_folder_path,
            feature_snake_name + "_feature",
            "CMakeLists.txt",
        )

    return generation_dict


def get_lazy_loading_fields(entity_mappable_with: str, entities_by_name: dict) -> list:
    entity_mappable_with_pascal = stringcase.pascalcase(entity_mappable_with)

    # create a list of foreign entities
    lazy_loading_fields = []
    for field in entities_by_name[entity_mappable_with_pascal]["fields"]:
        field_type = field["type"]
        field_name = field["name"]
        if tools.is_unique_foreign_entity(field_type, entities_by_name):
            lazy_loading_fields += [stringcase.pascalcase(field_name)]
        if tools.is_list_foreign_entity(field_type, entities_by_name):
            lazy_loading_fields += [stringcase.pascalcase(field_name)]

    return lazy_loading_fields


def generate_common_cmakelists(
    root_path: str,
    feature_by_name: dict,
    common_cmake_folder_path: str,
    files_to_be_generated: dict[str, bool],
):
    # generate common cmakelists.txt

    relative_common_cmakelists_file = os.path.join(
        common_cmake_folder_path, "CMakeLists.txt"
    )
    common_cmakelists_file = os.path.join(root_path, relative_common_cmakelists_file)
    if files_to_be_generated.get(relative_common_cmakelists_file, False):
        ## After the loop, write the list of features folders to the common cmakelists.txt
        with open(common_cmakelists_file, "w") as fh:
            for feature_name, _ in feature_by_name.items():
                fh.write(
                    f"add_subdirectory({stringcase.snakecase(feature_name)}_feature)"
                    + "\n"
                )
            print(f"Successfully wrote file {common_cmakelists_file}")


def generate_feature_export_file(
    root_path: str,
    feature: dict,
    application_name: str,
    files_to_be_generated: dict[str, bool],
):
    template_env = Environment(loader=FileSystemLoader("templates/application"))
    dto_cmakelists_template = template_env.get_template("export_template.jinja2")

    export_header_file = feature["export_header_file"]
    export = feature["export"]

    if not files_to_be_generated.get(export_header_file, False):
        return

    application_uppercase_name = stringcase.uppercase(application_name)

    feature_snake_name = feature["feature_snake_name"]
    feature_uppercase_name = stringcase.uppercase(feature_snake_name)

    rendered_template = dto_cmakelists_template.render(
        application_uppercase_name=application_uppercase_name,
        feature_uppercase_name=feature_uppercase_name,
        export=export,
    )

    export_header_file = os.path.join(root_path, export_header_file)

    # Create the directory if it does not exist
    os.makedirs(os.path.dirname(export_header_file), exist_ok=True)

    with open(export_header_file, "w") as fh:
        fh.write(rendered_template)
        print(f"Successfully wrote file {export_header_file}")


def generate_handler_cmakelists(
    root_path: str,
    feature: dict,
    application_name: str,
    files_to_be_generated: dict[str, bool],
):
    # generate these DTO's cmakelists.txt

    template_env = Environment(loader=FileSystemLoader("templates/application"))
    dto_cmakelists_template = template_env.get_template("cmakelists_template.jinja2")

    dto_cmakelists_file = feature["cmakelists_file"]

    if not files_to_be_generated.get(dto_cmakelists_file, False):
        return

    files = feature["handler_files"] + [feature["export_header_file"]]

    dto_cmakelists_file = os.path.join(root_path, dto_cmakelists_file)

    ## Convert the file path to be relative to the directory of the cmakelists
    relative_generated_files = []
    for file_path in files:
        relative_generated_file = os.path.relpath(
            os.path.join(root_path, file_path), os.path.dirname(dto_cmakelists_file)
        )
        relative_generated_files.append(relative_generated_file.replace("\\", "/"))

    feature_snake_name = feature["feature_snake_name"]
    feature_spinal_name = stringcase.spinalcase(feature_snake_name)

    rendered_template = dto_cmakelists_template.render(
        feature_snake_name=feature_snake_name,
        feature_spinal_name=feature_spinal_name,
        feature_uppercase_name=stringcase.uppercase(feature_snake_name),
        files=relative_generated_files,
        application_spinalcase_name=stringcase.spinalcase(application_name),
        application_uppercase_name=stringcase.uppercase(application_name),
    )

    # Create the directory if it does not exist
    os.makedirs(os.path.dirname(dto_cmakelists_file), exist_ok=True)

    with open(dto_cmakelists_file, "w") as fh:
        fh.write(rendered_template)
        print(f"Successfully wrote file {dto_cmakelists_file}")


def generate_crud_handler(
    root_path: str,
    handler: dict,
    feature: dict,
    files_to_be_generated: dict[str, bool],
    application_cpp_domain_name: str,
    uncrustify_config_file: str,
):
    for file_path in handler["files"]:
        if not files_to_be_generated.get(file_path, False):
            continue

        template_env = Environment(
            loader=FileSystemLoader("templates/application/CRUD")
        )
        template = template_env.get_template(
            handler["templates"][handler["files"].index(file_path)]
        )
        file_path = os.path.join(root_path, file_path)

        # Create the directory if it does not exist
        os.makedirs(os.path.dirname(file_path), exist_ok=True)

        with open(file_path, "w") as f:
            f.write(
                template.render(
                    feature_pascal_name=feature["feature_pascal_name"],
                    feature_snake_name=feature["feature_snake_name"],
                    export=handler["export"],
                    export_header=handler["export_header"],
                    entity_mappable_with_pascal=handler["entity_mappable_with_pascal"],
                    entity_mappable_with_snake=handler["entity_mappable_with_snake"],
                    entity_mappable_with_camel=handler["entity_mappable_with_camel"],
                    lazy_load_pascal_fields=handler.get("lazy_load_pascal_fields", []),
                    has_owner=handler.get("has_owner", False),
                    owner_name_snake=handler.get("owner_name_snake", ""),
                    owner_name_pascal=handler.get("owner_name_pascal", ""),
                    owner_name_camel=handler.get("owner_name_camel", ""),
                    owner_field_name_snake=handler.get("owner_field_name_snake", ""),
                    owner_field_name_pascal=handler.get("owner_field_name_pascal", ""),
                    owner_field_name_camel=handler.get("owner_field_name_camel", ""),
                    application_cpp_domain_name=application_cpp_domain_name,
                    owner_field_is_list=handler.get("owner_field_is_list", False),
                    owner_field_is_ordered=handler.get("owner_field_is_ordered", False),
                )
            )
            print(f"Successfully wrote file {file_path}")

        # if uncrustify_config_file:
        #     uncrustify.run_uncrustify(file_path, uncrustify_config_file)
        clang_format.run_clang_format(file_path)


def generate_custom_command_handler(
    root_path: str,
    handler: dict,
    feature: dict,
    files_to_be_generated: dict[str, bool],
    application_cpp_domain_name: str,
    uncrustify_config_file: str,
):
    for file_path in handler["files"]:
        if not files_to_be_generated.get(file_path, False):
            continue

        template_env = Environment(
            loader=FileSystemLoader("templates/application/custom")
        )
        template = template_env.get_template(
            handler["templates"][handler["files"].index(file_path)]
        )
        file_path = os.path.join(root_path, file_path)

        # Create the directory if it does not exist
        os.makedirs(os.path.dirname(file_path), exist_ok=True)

        with open(file_path, "w") as f:
            f.write(
                template.render(
                    feature_pascal_name=feature["feature_pascal_name"],
                    feature_snake_name=feature["feature_snake_name"],
                    command=handler,
                    export=handler["export"],
                    export_header=handler["export_header"],
                    validator_enabled=handler.get("validator", {}).get(
                        "enabled", False
                    ),
                    application_cpp_domain_name=application_cpp_domain_name,
                )
            )
            print(f"Successfully wrote file {file_path}")

        # if uncrustify_config_file:
        #     uncrustify.run_uncrustify(file_path, uncrustify_config_file)
        clang_format.run_clang_format(file_path)


def generate_custom_query_handler(
    root_path: str,
    handler: dict,
    feature: dict,
    files_to_be_generated: dict[str, bool],
    application_cpp_domain_name: str,
    uncrustify_config_file: str,
):
    for file_path in handler["files"]:
        if not files_to_be_generated.get(file_path, False):
            continue

        template_env = Environment(
            loader=FileSystemLoader("templates/application/custom")
        )
        template = template_env.get_template(
            handler["templates"][handler["files"].index(file_path)]
        )
        file_path = os.path.join(root_path, file_path)

        # Create the directory if it does not exist
        os.makedirs(os.path.dirname(file_path), exist_ok=True)

        with open(file_path, "w") as f:
            f.write(
                template.render(
                    feature_pascal_name=feature["feature_pascal_name"],
                    feature_snake_name=feature["feature_snake_name"],
                    feature_camel_name=feature["feature_camel_name"],
                    query=handler,
                    export=handler["export"],
                    export_header=handler["export_header"],
                    validator_enabled=handler.get("validator", {}).get(
                        "enabled", False
                    ),
                    application_cpp_domain_name=application_cpp_domain_name,
                )
            )
            print(f"Successfully wrote file {file_path}")

        # if uncrustify_config_file:
        #     uncrustify.run_uncrustify(file_path, uncrustify_config_file)
        clang_format.run_clang_format(file_path)


def generate_application_files(
    root_path: str,
    manifest_file: str,
    files_to_be_generated: dict[str, bool] = {},
    uncrustify_config_file: str = "",
):
    with open(manifest_file, "r") as stream:
        try:
            manifest_data = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)
            return

    application_name = manifest_data.get("global", {}).get(
        "application_name", "example"
    )

    application_data = manifest_data.get("application", [])
    feature_list = application_data.get("features", [])
    common_cmake_folder_path = application_data.get("common_cmake_folder_path", "")

    # Organize feature_list by name for easier lookup
    feature_by_name = {feature["name"]: feature for feature in feature_list}

    global_data = manifest_data.get("global", [])
    application_cpp_domain_name = global_data.get(
        "application_cpp_domain_name", "Undefined"
    )

    entities_data = manifest_data.get("entities", [])
    entities_list = entities_data.get("list", [])
    # remove entities that are not to be generated
    entities_list = [entity for entity in entities_list]

    # Organize entities by name for easier lookup
    entities_by_name = {entity["name"]: entity for entity in entities_list}

    for _, feature in get_generation_dict(
        common_cmake_folder_path, application_name, feature_by_name, entities_by_name
    ).items():
        # generate crud handlers
        for handler in feature["crud_handlers"].values():
            generate_crud_handler(
                root_path,
                handler,
                feature,
                files_to_be_generated,
                application_cpp_domain_name,
                uncrustify_config_file,
            )

        # generate custom handlers
        if feature.get("custom_handlers", None):
            if feature["custom_handlers"].get("commands", None):
                for command_handler in feature["custom_handlers"]["commands"]:
                    generate_custom_command_handler(
                        root_path,
                        command_handler,
                        feature,
                        files_to_be_generated,
                        application_cpp_domain_name,
                        uncrustify_config_file,
                    )
            if feature["custom_handlers"].get("queries", None):
                for query_handler in feature["custom_handlers"]["queries"]:
                    generate_custom_query_handler(
                        root_path,
                        query_handler,
                        feature,
                        files_to_be_generated,
                        application_cpp_domain_name,
                        uncrustify_config_file,
                    )

        # generate handler cmakelists.txt
        generate_handler_cmakelists(
            root_path, feature, application_name, files_to_be_generated
        )

        generate_feature_export_file(
            root_path, feature, application_name, files_to_be_generated
        )

    # generate common cmakelists.txt
    generate_common_cmakelists(
        root_path, feature_by_name, common_cmake_folder_path, files_to_be_generated
    )


def get_files_to_be_generated(
    manifest_file: str, files_to_be_generated: dict[str, bool] = {}
) -> list[str]:
    """
    Get the list of files that need to be generated based on the manifest file
    """
    # Read the manifest file
    with open(manifest_file, "r") as stream:
        try:
            manifest_data = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)
            return

    application_name = manifest_data.get("global", {}).get(
        "application_name", "example"
    )
    application_name = stringcase.spinalcase(application_name)

    application_data = manifest_data.get("application", [])
    feature_list = application_data.get("features", [])
    common_cmake_folder_path = application_data.get("common_cmake_folder_path", "")

    # Organize feature_list by name for easier lookup
    feature_by_name = {feature["name"]: feature for feature in feature_list}

    entities_data = manifest_data.get("entities", [])
    entities_list = entities_data.get("list", [])
    # remove entities that are not to be generated
    entities_list = [entity for entity in entities_list if entity.get("generate", True)]

    # Organize entities by name for easier lookup
    entities_by_name = {entity["name"]: entity for entity in entities_list}

    files = []
    for _, feature in get_generation_dict(
        common_cmake_folder_path, application_name, feature_by_name, entities_by_name
    ).items():
        files += feature["handler_files"]

        common_cmake_file = feature["cmakelists_file"]
        files.append(common_cmake_file)
        export_header_file = feature["export_header_file"]
        files.append(export_header_file)

    # # add CMakelists.txt:
    common_cmake_file = os.path.join(common_cmake_folder_path, "CMakeLists.txt")
    files.append(common_cmake_file)

    # strip from files if the value in files_to_be_generated is False
    if files_to_be_generated:
        for path, generate in files_to_be_generated.items():
            if not generate:
                files.remove(path)

    return files


# generate the files into the preview folder
def preview_application_files(
    root_path: str,
    manifest_file: str,
    files_to_be_generated: dict[str, bool] = {},
    uncrustify_config_file: str = "",
):
    manifest_preview_file = "temp/manifest_preview.yaml"

    # make a copy of the manifest file into temp/manifest_preview.yaml
    shutil.copy(manifest_file, manifest_preview_file)

    # modify the manifest file to generate the files into the preview folder
    with open(manifest_preview_file, "r") as fh:
        manifest = yaml.safe_load(fh)

    # remove .. from the path
    manifest["application"]["common_cmake_folder_path"] = manifest["application"][
        "common_cmake_folder_path"
    ].replace("..", "")

    # write the modified manifest file
    with open(manifest_preview_file, "w") as fh:
        yaml.dump(manifest, fh)

    root_path = os.path.join(root_path, "qleany_preview")

    # remove .. from the path
    if files_to_be_generated:
        preview_files_to_be_generated = {}
        for path, value in files_to_be_generated.items():
            preview_files_to_be_generated[path.replace("..", "")] = value

        generate_application_files(
            root_path,
            manifest_preview_file,
            preview_files_to_be_generated,
            uncrustify_config_file,
        )

    else:
        generate_application_files(
            root_path, manifest_preview_file, {}, uncrustify_config_file
        )


# Main execution
if __name__ == "__main__":
    full_path = Path(__file__).resolve().parent

    # Add the current directory to the path so that we can import the generated files
    sys.path.append(full_path)

    # Set the current directory to the generator directory
    os.chdir(full_path)

    if len(sys.argv) > 1:
        manifest_arg = sys.argv[1]
        if manifest_arg.endswith(".yaml") or manifest_arg.endswith(".yml"):
            manifest_file = manifest_arg
            root_path = Path(manifest_file).parent

            if len(sys.argv) > 2 and sys.argv[2] == "--preview":
                preview_application_files(root_path, manifest_file)
            else:
                # generate the files
                generate_application_files(root_path, manifest_file)

        else:
            print("Error: Manifest file must be named 'qleany.yaml' or 'qleany.yml'")
    else:
        print("Error: Please provide the manifest file as an argument")



import stringcase
import copy


def is_unique_foreign_entity(field_type: str, entities_by_name: dict) -> bool:
    for entity_name, _ in entities_by_name.items():
        if entity_name == field_type:
            return True

    return False


def is_list_foreign_entity(field_type: str, entities_by_name: dict) -> bool:
    if "<" not in field_type:
        return False

    type = field_type.split("<")[1].split(">")[0].strip()

    for entity_name, _ in entities_by_name.items():
        if entity_name == type:
            return True

    return False


def does_entity_have_relation_fields(entity_name: str, entities_by_name: dict) -> bool:
    entity = entities_by_name.get(entity_name, None)
    if entity is None:
        return False

    fields = entity["fields"]
    for field in fields:
        field_type = field["type"]
        if is_unique_foreign_entity(
            field_type, entities_by_name
        ) or is_list_foreign_entity(field_type, entities_by_name):
            return True

    return False


def get_entity_from_foreign_field_type(field_type: str, entities_by_name: dict) -> str:
    if "<" not in field_type:
        return field_type

    type = field_type.split("<")[1].split(">")[0].strip()

    for entity_name in entities_by_name:
        if entity_name == type:
            return entity_name

    return ""


def determine_owner(entity_name: str, entities_by_name: dict) -> dict:
    owner_dict = {}
    for possible_owner_name, entity in entities_by_name.items():
        for field in entity["fields"]:
            if field["type"] == entity_name or field["type"] == f"QList<{entity_name}>":
                if field.get("strong", False):
                    owner_dict["name"] = possible_owner_name
                    owner_dict["field"] = field["name"]
                    owner_dict["ordered"] = field.get("ordered", False)
                    owner_dict["is_list"] = field["type"] == f"QList<{entity_name}>"
                    return owner_dict

    return owner_dict


def get_fields_with_foreign_entities(
    fields: list, entities_by_name: dict, target_entity: str = ""
) -> list:
    if fields is None:
        return []

    # make a deep copy of the fields
    fields = copy.deepcopy(fields)

    # get recursive fields from parent

    parent_fields = []
    entity_parent = entities_by_name.get(target_entity, {}).get("parent", "")

    while entity_parent:
        parent_fields = (
            get_entity_fields(entity_parent, entities_by_name) + parent_fields
        )
        entity_parent = entities_by_name.get(entity_parent, {}).get("parent", "")

    fields = parent_fields + fields

    # add fields with foreign entities
    for field in fields:
        field["pascal_name"] = stringcase.pascalcase(field["name"])

        if is_unique_foreign_entity(
            field["type"], entities_by_name
        ) or is_list_foreign_entity(field["type"], entities_by_name):
            field["is_foreign"] = True

            # get foreign entity name
            foreign_entity_name = get_entity_from_foreign_field_type(
                field["type"], entities_by_name
            )
            field["foreign_dto_type"] = f"{foreign_entity_name}DTO"
            field["entity_type"] = field["type"]
            field["type"] = (
                f"{foreign_entity_name}DTO"
                if field["type"].count(">") == 0
                else f"QList<{foreign_entity_name}DTO>"
            )

        else:
            field["is_foreign"] = False

    return fields


def get_fields_without_foreign_entities(
    fields: list, entities_by_name: dict, target_entity: str = ""
) -> list:
    # make a deep copy of the fields
    fields = copy.deepcopy(fields)

    # get recursive fields from parent

    parent_fields = []
    entity_parent = entities_by_name.get(target_entity, {}).get("parent", "")

    while entity_parent:
        parent_fields = (
            get_entity_fields(entity_parent, entities_by_name) + parent_fields
        )
        entity_parent = entities_by_name.get(entity_parent, {}).get("parent", "")

    fields = parent_fields + fields

    # add fields without foreign entities
    fields_without_foreign = []
    for field in fields:
        field["pascal_name"] = stringcase.pascalcase(field["name"])
        field["camel_name"] = stringcase.camelcase(field["name"])

        if is_unique_foreign_entity(
            field["type"], entities_by_name
        ) or is_list_foreign_entity(field["type"], entities_by_name):
            continue

        else:
            field["default_value_for_qml"] = get_default_value_for_qml(field["type"])
            field["is_foreign"] = False
            fields_without_foreign.append(field)

    return fields_without_foreign

def get_default_value_for_qml(type: str):
    if type == "int":
        return "0"
    elif type == "double":
        return "0.0"
    elif type == "bool":
        return "false"
    elif type == "QString":
        return '"example"'
    elif type == "QDate":
        return "new Date()"
    elif type == "QTime":
        return "new Date()"
    elif type == "QDateTime":
        return "new Date()"
    elif type == "QUrl":
        return '""'
    elif type == "QUuid":
        return '""'
    elif type == "QByteArray":
        return '""'
    elif type == "QVariant":
        return '""'
    elif type == "QList<int>":
        return "[]"
    elif type == "QList<double>":
        return "[]"
    elif type == "QList<bool>":
        return "[]"
    elif type == "QList<QString>":
        return "[]"
    elif type == "QList<QDate>":
        return "[]"
    elif type == "QList<QTime>":
        return "[]"
    elif type == "QList<QDateTime>":
        return "[]"
    elif type == "QList<QUrl>":
        return "[]"
    elif type == "QList<QByteArray>":
        return "[]"
    elif type == "QList<QVariant>":
        return "[]"
    else:
        return "null"
    


def get_only_fields_with_foreign_entities(
    fields: list, entities_by_name: dict, target_entity: str = ""
) -> list:
    # make a deep copy of the fields
    fields = copy.deepcopy(fields)

    # get recursive fields from parent

    parent_fields = []
    entity_parent = entities_by_name.get(target_entity, {}).get("parent", "")

    while entity_parent:
        parent_fields = (
            get_entity_fields(entity_parent, entities_by_name) + parent_fields
        )
        entity_parent = entities_by_name.get(entity_parent, {}).get("parent", "")

    fields = parent_fields + fields

    # add fields without foreign entities
    fields_with_foreign = []
    for field in fields:
        field["pascal_name"] = stringcase.pascalcase(field["name"])

        if is_unique_foreign_entity(
            field["type"], entities_by_name
        ) or is_list_foreign_entity(field["type"], entities_by_name):
            field["is_foreign"] = True

            # get foreign entity name
            foreign_entity_name = get_entity_from_foreign_field_type(
                field["type"], entities_by_name
            )
            field["foreign_dto_type"] = f"{foreign_entity_name}DTO"
            field["entity_type"] = field["type"]
            field["type"] = (
                f"{foreign_entity_name}DTO"
                if field["type"].count(">") == 0
                else f"QList<{foreign_entity_name}DTO>"
            )
            fields_with_foreign.append(field)

        else:
            continue

    return fields_with_foreign


def get_entity_fields(entity_name: str, entities_by_name: dict) -> list:
    if entity_name == "EntityBase":
        return [{"name": "id", "type": "int", "pascal_name": "Id", "is_foreign": False}]

    entity_data = entities_by_name[entity_name]
    fields = entity_data["fields"]
    return fields



from jinja2 import Environment, FileSystemLoader
import yaml
import os
import sys
import stringcase
import shutil
import uncrustify
import clang_format
from pathlib import Path
from copy import deepcopy
import generation_dict_tools as tools


def generate_export_header_file(
    root_path: str,
    path: str,
    application_name: str,
    export: str,
    export_header_file: str,
    layer_name: str,
    files_to_be_generated: dict[str, bool],
):
    # generate the export header file

    template_env = Environment(loader=FileSystemLoader("templates/repositories"))
    export_header_template = template_env.get_template("export_template.jinja2")

    relative_export_header_file = os.path.join(path, export_header_file)

    if not files_to_be_generated.get(relative_export_header_file, False):
        return

    export_header_file = os.path.join(root_path, relative_export_header_file)

    rendered_template = export_header_template.render(
        application_uppercase_name=stringcase.uppercase(application_name),
        export=export,
        layer_uppercase_name=stringcase.uppercase(layer_name),
    )

    # Create the directory if it does not exist
    os.makedirs(os.path.dirname(export_header_file), exist_ok=True)

    with open(export_header_file, "w") as fh:
        fh.write(rendered_template)
        print(f"Successfully wrote file {export_header_file}")


def generate_cmakelists(
    root_path: str,
    path: str,
    application_name: str,
    export_header_file: str,
    files_to_be_generated: dict[str, bool],
):
    # generate the cmakelists.txt

    template_env = Environment(loader=FileSystemLoader("templates/repositories"))
    cmakelists_template = template_env.get_template("cmakelists_template.jinja2")

    relative_cmakelists_file = os.path.join(path, "CMakeLists.txt")

    if not files_to_be_generated.get(relative_cmakelists_file, False):
        return

    cmakelists_file = os.path.join(root_path, relative_cmakelists_file)

    rendered_template = cmakelists_template.render(
        application_spinalcase_name=stringcase.spinalcase(application_name),
        application_uppercase_name=stringcase.uppercase(application_name),
        export_header_file=export_header_file,
    )

    # Create the directory if it does not exist
    os.makedirs(os.path.dirname(cmakelists_file), exist_ok=True)

    with open(cmakelists_file, "w") as fh:
        fh.write(rendered_template)
        print(f"Successfully wrote file {cmakelists_file}")


def generate_contracts_cmakelists(
    root_path: str,
    path: str,
    application_name: str,
    export_header_file: str,
    files_to_be_generated: dict[str, bool],
):
    # generate the cmakelists.txt

    template_env = Environment(loader=FileSystemLoader("templates/repositories"))
    cmakelists_template = template_env.get_template(
        "contracts_cmakelists_template.jinja2"
    )

    relative_cmakelists_file = os.path.join(path, "CMakeLists.txt")

    if not files_to_be_generated.get(relative_cmakelists_file, False):
        return

    cmakelists_file = os.path.join(root_path, relative_cmakelists_file)

    rendered_template = cmakelists_template.render(
        application_spinalcase_name=stringcase.spinalcase(application_name),
        application_uppercase_name=stringcase.uppercase(application_name),
        export_header_file=export_header_file,
    )

    # Create the directory if it does not exist
    os.makedirs(os.path.dirname(cmakelists_file), exist_ok=True)

    with open(cmakelists_file, "w") as fh:
        fh.write(rendered_template)
        print(f"Successfully wrote file {cmakelists_file}")


def generate_repository_files(
    root_path: str,
    manifest_file,
    files_to_be_generated: dict[str, bool] = None,
    uncrustify_config_file: str = None,
):
    with open(manifest_file, "r") as stream:
        try:
            manifest_data = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)
            return

    repositories_data = manifest_data.get("repositories", [])
    repositories_list = repositories_data.get("list", [])
    export = repositories_data.get("export", "")
    export_header_file = repositories_data.get(
        "export_header_file", "repositories_export.h"
    )
    repository_path = repositories_data.get("repository_folder_path", ".")
    base_path = repositories_data.get("base_folder_path", ".")
    interface_path = repositories_data.get("interface_path", ".")

    global_data = manifest_data.get("global", [])
    application_name = global_data.get("application_name", "example")
    application_cpp_domain_name = global_data.get(
        "application_cpp_domain_name", "Undefined"
    )

    template_env = Environment(loader=FileSystemLoader("templates/repositories"))
    repo_template = template_env.get_template("repository_template.h.jinja2")
    repo_cpp_template = template_env.get_template("repository_template.cpp.jinja2")
    interface_template = template_env.get_template(
        "interface_repository_template.h.jinja2"
    )

    contracts_data = manifest_data.get("contracts", [])
    contracts_export = contracts_data.get("export", "")
    contracts_export_header_file = contracts_data.get("export_header_file", "")
    inverted_app_domain = contracts_data.get("inverted_app_domain", "com.example")
    contracts_folder_path = contracts_data.get("folder_path", ".")

    entities_data = manifest_data.get("entities", [])
    entities_list = entities_data.get("list", [])

    # Organize entities by name for easier lookup
    entities_by_name = {entity["name"]: entity for entity in entities_list}

    generated_files = []
    interface_generated_files = []

    all_repository_files = []
    all_interface_files = []

    for repository in repositories_list:
        name = repository["entity_name"]
        pascal_name = stringcase.pascalcase(name)
        snake_name = stringcase.snakecase(name)
        camel_name = stringcase.camelcase(name)
        generate_lazy_loaders = repository.get("lazy_loaders", True)

        # create a dict of foreign entities for this repository
        foreign_entities = {}
        for field in entities_by_name[name]["fields"]:
            field_type = field["type"]
            field_name = field["name"]
            if tools.is_unique_foreign_entity(field_type, entities_by_name):
                foreign_entities[f"{field_type}__{field_name}"] = deepcopy(
                    entities_by_name[field_type]
                )
            elif tools.is_list_foreign_entity(field_type, entities_by_name):
                foreign_entities[f"{field_type}__{field_name}"] = deepcopy(
                    entities_by_name[field_type.split("<")[1].split(">")[0].strip()]
                )

        foreign_entities = dict(sorted(foreign_entities.items()))

        # add helpful keys to foreign entities
        for key, value in foreign_entities.items():
            field_type = key.split("__")[0]
            field_name = key.split("__")[1]
            value["is_list"] = field_type.split("<")[
                0
            ] == "QList" and tools.is_list_foreign_entity(field_type, entities_by_name)
            if value["is_list"]:
                value["type_camel_name"] = stringcase.camelcase(
                    field_type.split("<")[1].split(">")[0].strip()
                )
            else:
                value["type_camel_name"] = stringcase.camelcase(field_type)

            if value["is_list"]:
                value["type_name_only"] = stringcase.pascalcase(
                    field_type.split("<")[1].split(">")[0].strip()
                )
            else:
                value["type_name_only"] = stringcase.pascalcase(field_type)

            if value["is_list"]:
                value["type_pascal_name"] = stringcase.pascalcase(
                    field_type.split("<")[1].split(">")[0].strip()
                )
            else:
                value["type_pascal_name"] = stringcase.pascalcase(field_type)

            if value["is_list"]:
                value["type_snake_name"] = stringcase.snakecase(
                    field_type.split("<")[1].split(">")[0].strip()
                )
            else:
                value["type_snake_name"] = stringcase.snakecase(field_type)

            value["related_field_name"] = field_name
            value["related_field_pascal_name"] = stringcase.pascalcase(field_name)

        foreign_repository_constructor_arguments = []
        if generate_lazy_loaders:
            for key, value in foreign_entities.items():
                new_constructor_argument = f"Interface{value['type_pascal_name']}Repository *{value['type_camel_name']}Repository"
                if (
                    new_constructor_argument
                    not in foreign_repository_constructor_arguments
                ):
                    foreign_repository_constructor_arguments.append(
                        new_constructor_argument
                    )

        foreign_repository_constructor_arguments_string = ", ".join(
            foreign_repository_constructor_arguments
        )
        foreign_repository_constructor_arguments_string = (
            ", " + foreign_repository_constructor_arguments_string
            if foreign_repository_constructor_arguments
            else foreign_repository_constructor_arguments_string
        )

        loader_private_member_list = []
        if generate_lazy_loaders:
            for key, value in foreign_entities.items():
                loader_private_member_list.append(
                    f"Interface{value['type_pascal_name']}Repository *m_{value['type_camel_name']}Repository;"
                )
        # remove duplicates :
        loader_private_member_list = list(dict.fromkeys(loader_private_member_list))

        foreign_repository_header_list = []
        if generate_lazy_loaders:
            for key, value in foreign_entities.items():
                foreign_repository_header_list.append(
                    f"\"repository/interface_{value['type_snake_name']}_repository.h\""
                )
        # remove duplicates :
        foreign_repository_header_list = list(
            dict.fromkeys(foreign_repository_header_list)
        )

        # loader functions like     Domain::Book::ChaptersLoader fetchChaptersLoader();

        loader_function_list = []
        if generate_lazy_loaders:
            for key, value in foreign_entities.items():
                loader_function_list.append(
                    f"{application_cpp_domain_name}::Domain::{name}::{value['related_field_pascal_name']}Loader fetch{value['related_field_pascal_name']}Loader() override;"
                )

        # Create .h file
        rendered_template = repo_template.render(
            name=name,
            pascal_name=pascal_name,
            snake_name=snake_name,
            camel_name=camel_name,
            foreign_repository_constructor_arguments_string=foreign_repository_constructor_arguments_string,
            loader_private_member_list=loader_private_member_list,
            loader_function_list=loader_function_list,
            export=export,
            export_header_file=export_header_file,
            foreign_repository_header_list=foreign_repository_header_list,
            application_cpp_domain_name=application_cpp_domain_name,
        )
        output_file = os.path.join(repository_path, f"{snake_name}_repository.h")
        all_repository_files.append(os.path.join(root_path, output_file))

        if files_to_be_generated.get(output_file, False):
            output_file = os.path.join(root_path, output_file)

            os.makedirs(os.path.dirname(output_file), exist_ok=True)
            with open(output_file, "w") as fh:
                fh.write(rendered_template)
                print(f"Successfully wrote file {output_file}")
            generated_files.append(output_file)

            # if uncrustify_config_file:
            #     uncrustify.run_uncrustify(output_file, uncrustify_config_file)
            clang_format.run_clang_format(output_file)

        # prepare the fields init values

        fields_init_values = [
            f"m_{value['type_camel_name']}Repository({value['type_camel_name']}Repository)"
            for key, value in foreign_entities.items()
        ]
        # remove duplicates :
        fields_init_values = list(dict.fromkeys(fields_init_values))

        fields_init_values = ", ".join(fields_init_values)
        fields_init_values = (
            ", " + fields_init_values if fields_init_values else fields_init_values
        )

        # Create .cpp file
        rendered_template = repo_cpp_template.render(
            name=name,
            pascal_name=pascal_name,
            snake_name=snake_name,
            camel_name=camel_name,
            foreign_entities=foreign_entities,
            foreign_repository_constructor_arguments_string=foreign_repository_constructor_arguments_string,
            fields_init_values=fields_init_values,
            application_cpp_domain_name=application_cpp_domain_name,
        )
        output_file = os.path.join(repository_path, f"{snake_name}_repository.cpp")
        all_repository_files.append(os.path.join(root_path, output_file))

        if files_to_be_generated.get(output_file, False):
            output_file = os.path.join(root_path, output_file)

            os.makedirs(os.path.dirname(output_file), exist_ok=True)
            with open(output_file, "w") as fh:
                fh.write(rendered_template)
                print(f"Successfully wrote file {output_file}")
            generated_files.append(output_file)

            # if uncrustify_config_file:
            #     uncrustify.run_uncrustify(output_file, uncrustify_config_file)
            clang_format.run_clang_format(output_file)

        # loader functions like     Domain::Book::ChaptersLoader fetchChaptersLoader();

        loader_function_list_for_interface = []
        if generate_lazy_loaders:
            for key, value in foreign_entities.items():
                loader_function_list_for_interface.append(
                    f"virtual {application_cpp_domain_name}::Domain::{name}::{value['related_field_pascal_name']}Loader fetch{value['related_field_pascal_name']}Loader() = 0;"
                )

        # Create interface .h file
        rendered_template = interface_template.render(
            name=name,
            snake_name=snake_name,
            camel_name=camel_name,
            inverted_app_domain=inverted_app_domain,
            contracts_export=contracts_export,
            contracts_export_header_file=contracts_export_header_file,
            loader_function_list_for_interface=loader_function_list_for_interface,
            application_cpp_domain_name=application_cpp_domain_name,
        )
        output_file = os.path.join(
            interface_path, f"interface_{snake_name}_repository.h"
        )
        all_interface_files.append(os.path.join(root_path, output_file))

        if files_to_be_generated.get(output_file, False):
            output_file = os.path.join(root_path, output_file)

            os.makedirs(os.path.dirname(output_file), exist_ok=True)
            with open(output_file, "w") as fh:
                fh.write(rendered_template)
                print(f"Successfully wrote file {output_file}")
            interface_generated_files.append(output_file)

            # if uncrustify_config_file:
            #     uncrustify.run_uncrustify(output_file, uncrustify_config_file)
            clang_format.run_clang_format(output_file)

    # write the repository cmake list file

    repository_cmake_template = template_env.get_template("repositories.cmake.jinja2")
    relative_cmake_output_file = os.path.join(base_path, "repositories.cmake")
    cmake_output_file = os.path.join(root_path, relative_cmake_output_file)

    if files_to_be_generated.get(relative_cmake_output_file, False):
        repositories = []
        for repository in all_repository_files:
            repository_header = f"{repository}"
            relative_path = os.path.relpath(
                repository_header, os.path.dirname(cmake_output_file)
            )
            repositories.append(relative_path.replace("\\", "/"))

        # Create the directory if it does not exist
        os.makedirs(os.path.dirname(cmake_output_file), exist_ok=True)

        rendered_template = repository_cmake_template.render(
            repositories=repositories,
        )

        with open(cmake_output_file, "w") as fh:
            fh.write(rendered_template)
            print(f"Successfully wrote file {cmake_output_file}")

    # write the interface list file

    repository_interface_cmake_template = template_env.get_template(
        "repositories.cmake.jinja2"
    )
    relative_repository_interface_cmake_output_file = os.path.join(
        contracts_folder_path, "repository_interfaces.cmake"
    )
    repository_interface_cmake_output_file = os.path.join(
        root_path, relative_repository_interface_cmake_output_file
    )

    if files_to_be_generated.get(
        relative_repository_interface_cmake_output_file, False
    ):
        repositories = []
        for repository_interface in all_interface_files:
            repository_interface_header = f"{repository_interface}"
            relative_path = os.path.relpath(
                repository_interface_header,
                os.path.dirname(repository_interface_cmake_output_file),
            )
            repositories.append(relative_path.replace("\\", "/"))

        # Create the directory if it does not exist
        os.makedirs(
            os.path.dirname(repository_interface_cmake_output_file), exist_ok=True
        )

        rendered_template = repository_interface_cmake_template.render(
            repositories=repositories,
        )

        with open(repository_interface_cmake_output_file, "w") as fh:
            fh.write(rendered_template)
            print(f"Successfully wrote file {repository_interface_cmake_output_file}")

    def determine_direct_children(entity: dict) -> list:
        """
        Determine the direct children of an entity
        """
        direct_children = []
        for field in entity["fields"]:
            field_type = field["type"]
            if tools.is_unique_foreign_entity(field_type, entities_by_name):
                direct_children.append(field_type)
            elif tools.is_list_foreign_entity(field_type, entities_by_name):
                direct_children.append(field_type.split("<")[1].split(">")[0].strip())

        # alphabetize the list
        direct_children.sort()

        return direct_children

    def determine_order_of_registering_of_entities(entities_list: list) -> list:
        """
        Determine the children for each entity and determine an order of entities so the result is a list of entities in the order they should be registered.
        """

        # Remove entities that are only for heritage
        entities_list = [
            entity
            for entity in entities_list
            if not entity.get("only_for_heritage", False)
        ]

        # Get a dict of children for each entity
        children_for_each_entity = {}
        for entity in entities_list:
            children_for_each_entity[entity["name"]] = determine_direct_children(entity)

        root_entities = []
        for entity in entities_list:
            is_root_entity = True
            for children in children_for_each_entity.values():
                if entity["name"] in children:
                    is_root_entity = False
                    break
            if is_root_entity:
                root_entities.append(entity["name"])

        # alphabetize the list
        root_entities.sort()

        class EntityNode:
            def __init__(self, name: str, children: [str]):
                self.name = name
                self.children = children
                self.parents = []
                self.registered = False
                self.is_root = False

            def is_common_child(self):
                return len(self.other_parents) > 0

            def add_child(self, child: str):
                self.children.append(child)

            def add_parent(self, parent: str):
                self.parents.append(parent)

        # create a list of Entity objects
        entities = []

        # add
        for entity in entities_list:
            entities.append(
                EntityNode(entity["name"], children_for_each_entity[entity["name"]])
            )

        # add the parents to each entity
        for entity in entities:
            for child in entity.children:
                for other_entity in entities:
                    if child == other_entity.name:
                        other_entity.add_parent(entity.name)

        # if the entity has no parent, it is a root entity
        for entity in entities:
            if not entity.parents:
                entity.is_root = True

        # create a list of root entities
        root_entities = []
        for entity in entities:
            if entity.is_root:
                root_entities.append(entity.name)

        # add the root entities to the ordered_entities list
        ordered_entities = []
        while len(ordered_entities) < len(root_entities):
            for entity in entities:
                if not entity.registered:
                    if entity.is_root:
                        ordered_entities.append(entity.name)
                        entity.registered = True

        # order the entities from the most root entity to the most leaf entity. Rule : a entity must be registered before its parents. And append only in parents are already appended.
        while len(ordered_entities) < len(entities):
            for entity in entities:
                if not entity.registered:
                    # check if all parents are already registered. entity.parents giving string instead of EntityNode
                    if all([parent in ordered_entities for parent in entity.parents]):
                        ordered_entities.append(entity.name)
                        entity.registered = True

        # reverse the order
        ordered_entities.reverse()

        return ordered_entities

    # write persistence_registration.h file

    persistence_registration_template = template_env.get_template(
        "persistence_registration.h.jinja2"
    )
    relative_output_file = os.path.join(base_path, "persistence_registration.h")
    output_file = os.path.join(root_path, relative_output_file)

    if files_to_be_generated.get(relative_output_file, False):
        repositories = []
        for repository in all_repository_files:
            repository_header = f"{repository}"
            relative_path = os.path.relpath(
                repository_header, os.path.dirname(output_file)
            )
            repositories.append(relative_path.replace("\\", "/"))

        # Create the directory if it does not exist
        os.makedirs(os.path.dirname(output_file), exist_ok=True)

        rendered_template = persistence_registration_template.render(
            repositories=repositories,
            export=export,
            export_header_file=export_header_file,
            application_cpp_domain_name=application_cpp_domain_name,
        )

        with open(output_file, "w") as fh:
            fh.write(rendered_template)
            print(f"Successfully wrote file {output_file}")

        clang_format.run_clang_format(output_file)

    # write persistence_registration.cpp file

    persistence_registration_template = template_env.get_template(
        "persistence_registration.cpp.jinja2"
    )
    relative_output_file = os.path.join(base_path, "persistence_registration.cpp")
    output_file = os.path.join(root_path, relative_output_file)

    if files_to_be_generated.get(relative_output_file, False):
        ordered_entities = determine_order_of_registering_of_entities(entities_list)

        # using the order given by ordered_entities, create a list of entities dict with entity_pascal_name, entity_camel_name and a list of children_entities, each having a child_camel_name
        entities = []
        for entity in ordered_entities:
            entity_pascal_name = stringcase.pascalcase(entity)
            entity_camel_name = stringcase.camelcase(entity)
            pre_children_entities = []
            for child in determine_direct_children(entities_by_name[entity]):
                if child not in pre_children_entities:
                    pre_children_entities.append(child)

            children_entities = []
            for child in pre_children_entities:
                children_entities.append(
                    {
                        "child_camel_name": stringcase.camelcase(child),
                    }
                )
            entities.append(
                {
                    "entity_pascal_name": entity_pascal_name,
                    "entity_camel_name": entity_camel_name,
                    "children_entities": children_entities,
                }
            )

        repositories = []
        repository_headers = []
        for repository in all_repository_files:
            repository_header = f"{repository}"
            relative_path = os.path.relpath(
                repository_header, os.path.dirname(output_file)
            )
            repositories.append(relative_path.replace("\\", "/"))
            if relative_path.endswith(".h"):
                repository_headers.append(relative_path.replace("\\", "/"))

        # Create the directory if it does not exist
        os.makedirs(os.path.dirname(output_file), exist_ok=True)

        rendered_template = persistence_registration_template.render(
            repository_headers=repository_headers,
            entities=entities,
            application_cpp_domain_name=application_cpp_domain_name,
        )

        with open(output_file, "w") as fh:
            fh.write(rendered_template)
            print(f"Successfully wrote file {output_file}")

        clang_format.run_clang_format(output_file)

    # write the CMakeLists.txt file

    generate_cmakelists(
        root_path,
        base_path,
        application_name,
        export_header_file,
        files_to_be_generated,
    )

    # write the interface CMakeLists.txt file

    generate_contracts_cmakelists(
        root_path,
        contracts_folder_path,
        application_name,
        contracts_export_header_file,
        files_to_be_generated,
    )

    # write the export header file

    generate_export_header_file(
        root_path,
        base_path,
        application_name,
        export,
        export_header_file,
        "persistence",
        files_to_be_generated,
    )

    generate_export_header_file(
        root_path,
        contracts_folder_path,
        application_name,
        contracts_export,
        contracts_export_header_file,
        "contracts",
        files_to_be_generated,
    )


def get_files_to_be_generated(
    manifest_file: str, files_to_be_generated: dict[str, bool] = None
) -> list[str]:
    """
    Get the list of files that need to be generated based on the manifest file
    """
    # Read the manifest file
    with open(manifest_file, "r") as fh:
        manifest = yaml.safe_load(fh)

    base_folder_path = manifest["repositories"]["base_folder_path"]
    repository_folder_path = manifest["repositories"]["repository_folder_path"]
    interface_folder_path = manifest["repositories"]["interface_path"]
    contracts_folder_path = manifest["contracts"]["folder_path"]

    # Get the list of files to be generated
    files = []
    for entity in manifest["repositories"]["list"]:
        entity_name = entity["entity_name"]
        files.append(
            os.path.join(
                repository_folder_path,
                f"{stringcase.snakecase(entity_name)}_repository.h",
            )
        )
        files.append(
            os.path.join(
                repository_folder_path,
                f"{stringcase.snakecase(entity_name)}_repository.cpp",
            )
        )
        files.append(
            os.path.join(
                interface_folder_path,
                f"interface_{stringcase.snakecase(entity_name)}_repository.h",
            )
        )

    # add repository cmake file:
    files.append(os.path.join(base_folder_path, "repositories.cmake"))

    # add persistence_registration file:
    files.append(os.path.join(base_folder_path, "persistence_registration.h"))
    files.append(os.path.join(base_folder_path, "persistence_registration.cpp"))

    # add CMakeLists.txt file:
    files.append(os.path.join(base_folder_path, "CMakeLists.txt"))

    # add contracts CMakeLists.txt file:
    files.append(os.path.join(contracts_folder_path, "CMakeLists.txt"))

    # add export header file:
    files.append(
        os.path.join(base_folder_path, manifest["repositories"]["export_header_file"])
    )

    # add contracts export header file:
    files.append(
        os.path.join(contracts_folder_path, manifest["contracts"]["export_header_file"])
    )

    # add interface cmake file:
    files.append(
        os.path.join(
            contracts_folder_path,
            "repository_interfaces.cmake",
        )
    )

    # strip from files if the value in files_to_be_generated is False
    if files_to_be_generated:
        for path, generate in files_to_be_generated.items():
            if not generate:
                files.remove(path)

    return files


# generate the files into the preview folder
def preview_repository_files(
    root_path: str,
    manifest_file: str,
    files_to_be_generated: dict[str, bool] = None,
    uncrustify_config_file: str = None,
):
    manifest_preview_file = "temp/manifest_preview.yaml"

    # make a copy of the manifest file into temp/manifest_preview.yaml
    shutil.copy(manifest_file, manifest_preview_file)

    # modify the manifest file to generate the files into the preview folder
    with open(manifest_preview_file, "r") as fh:
        manifest = yaml.safe_load(fh)

    # remove .. from the path and add preview before the folder name
    manifest["repositories"]["base_folder_path"] = manifest["repositories"][
        "base_folder_path"
    ].replace("..", "")
    manifest["repositories"]["repository_folder_path"] = manifest["repositories"][
        "repository_folder_path"
    ].replace("..", "")
    manifest["repositories"]["interface_path"] = manifest["repositories"][
        "interface_path"
    ].replace("..", "")

    # write the modified manifest file
    with open(manifest_preview_file, "w") as fh:
        yaml.dump(manifest, fh)

    root_path = os.path.join(root_path, "qleany_preview")

    # preprend preview/ to the file names in the dict files_to_be_generated and remove .. from the path
    if files_to_be_generated:
        preview_files_to_be_generated = {}
        for path, value in files_to_be_generated.items():
            preview_files_to_be_generated[path.replace("..", "")] = value

        generate_repository_files(
            root_path,
            manifest_preview_file,
            preview_files_to_be_generated,
            uncrustify_config_file,
        )
    else:
        generate_repository_files(
            root_path, manifest_preview_file, {}, uncrustify_config_file
        )


# Main execution
if __name__ == "__main__":
    full_path = Path(__file__).resolve().parent

    # Add the current directory to the path so that we can import the generated files
    sys.path.append(full_path)

    # Set the current directory to the generator directory
    os.chdir(full_path)

    if len(sys.argv) > 1:
        manifest_arg = sys.argv[1]
        if manifest_arg.endswith(".yaml") or manifest_arg.endswith(".yml"):
            manifest_file = manifest_arg
            root_path = Path(manifest_file).parent

            if len(sys.argv) > 2 and sys.argv[2] == "--preview":
                preview_repository_files(root_path, manifest_file)
            else:
                generate_repository_files(root_path, manifest_file)
        else:
            print("Error: Manifest file must be named 'qleany.yaml' or 'qleany.yml'")
    else:
        print("Error: Please provide the manifest file as an argument")



from jinja2 import Environment, FileSystemLoader
import yaml
import os
import sys
import stringcase
import shutil
import uncrustify
import clang_format
import generation_dict_tools as tools
from pathlib import Path


def _get_generation_dict(
    real_imports_folder_path: str,
    mock_imports_folder_path: str,
    feature_by_name: dict,
    entities_by_name: dict,
    has_undo_redo: bool,
    create_undo_and_redo_singles: bool,
    list_models: list,
    singles: list,
    application_name: str,
    application_cpp_domain_name: str,
) -> dict:
    # generating controller files

    generation_dict = {}
    generation_dict["controllers"] = {}
    for feature_name, feature_data in feature_by_name.items():
        feature_snake_name = stringcase.snakecase(feature_name)
        feature_pascal_name = stringcase.pascalcase(feature_name)
        feature_camel_name = stringcase.camelcase(feature_name)

        crud = feature_data.get("CRUD", {})
        crud_enabled = crud.get("enabled", False)
        get_enabled = crud.get("get", {}).get("enabled", False)
        get_with_details_enabled = crud.get("get_with_details", {}).get(
            "enabled", False
        )
        get_all_enabled = crud.get("get_all", {}).get("enabled", False)
        create_enabled = crud.get("create", {}).get("enabled", False)
        update_enabled = crud.get("update", {}).get("enabled", False)
        remove_enabled = crud.get("remove", {}).get("enabled", False)

        # custom commands
        custom_commands = []
        for command in feature_data.get("commands", []):
            dto_in = command.get("dto", {}).get("in", {})
            custom_commands.append(
                {
                    "camel_name": stringcase.camelcase(command["name"]),
                    "pascal_name": stringcase.pascalcase(command["name"]),
                    "snake_name": stringcase.snakecase(command["name"]),
                    "dto_in_enabled": dto_in.get("enabled", True),
                    "dto_in_pascal_type_prefix": stringcase.pascalcase(
                        dto_in.get("type_prefix", "undefined")
                    ),
                }
            )

        # custom queries
        custom_queries = []
        for query in feature_data.get("queries", []):
            dto_in = query.get("dto", {}).get("in", {})
            custom_queries.append(
                {
                    "camel_name": stringcase.camelcase(query["name"]),
                    "pascal_name": stringcase.pascalcase(query["name"]),
                    "snake_name": stringcase.snakecase(query["name"]),
                    "dto_in_enabled": dto_in.get("enabled", True),
                    "dto_in_pascal_type_prefix": stringcase.pascalcase(
                        dto_in.get("type_prefix", "undefined")
                    ),
                }
            )

        generation_dict["controllers"][feature_pascal_name] = {
            "mock_controller_file": os.path.join(
                mock_imports_folder_path,
                "Controllers",
                f"{feature_pascal_name}Controller.qml",
            ),
            "mock_signals_file": os.path.join(
                mock_imports_folder_path,
                "Controllers",
                f"{feature_pascal_name}Signals.qml",
            ),
            "mock_template_path": "QML/mock_imports/controllers/",
            "mock_template": "controller.qml.jinja2",
            "mock_signals_template": "signals.qml.jinja2",
            "real_controller_file": os.path.join(
                real_imports_folder_path,
                "controllers",
                f"foreign_{feature_snake_name}_controller.h",
            ),
            "real_template_path": "QML/real_imports/controllers/",
            "real_template": "foreign_controller.h.jinja2",
            "feature_pascal_name": feature_pascal_name,
            "feature_camel_name": feature_camel_name,
            "feature_snake_name": feature_snake_name,
            "crud": {
                "enabled": crud_enabled,
                "get": get_enabled,
                "get_with_details": get_with_details_enabled,
                "get_all": get_all_enabled,
                "create": create_enabled,
                "update": update_enabled,
                "remove": remove_enabled,
                "entity_has_relation_fields": tools.does_entity_have_relation_fields(
                    feature_pascal_name, entities_by_name
                ),
            },
            "custom_commands": custom_commands,
            "custom_queries": custom_queries,
        }

    # add mock_custom_functions to the generation_dict["controllers"][feature_pascal_name] dict
    for feature_name, feature_data in feature_by_name.items():
        feature_pascal_name = stringcase.pascalcase(feature_name)
        generation_dict["controllers"][feature_pascal_name][
            "mock_custom_functions"
        ] = []
        commands = feature_data.get("commands", [])
        if commands:
            for command in commands:
                generation_dict["controllers"][feature_pascal_name][
                    "mock_custom_functions"
                ] += [stringcase.camelcase(command["name"])]

        queries = feature_data.get("queries", [])
        if queries:
            for query in queries:
                generation_dict["controllers"][feature_pascal_name][
                    "mock_custom_functions"
                ] += [stringcase.camelcase(query["name"])]

    # add qmldir:
    controller_qmldir_file = os.path.join(
        mock_imports_folder_path, "Controllers", "qmldir"
    )

    generation_dict["controller_qmldir_file"] = controller_qmldir_file

    # add CMakelists.txt:
    controller_cmakelists_file = os.path.join(
        real_imports_folder_path, "controllers", "CMakeLists.txt"
    )
    generation_dict["controller_cmakelists_file"] = controller_cmakelists_file

    # add "mock_controller_file" and "real_controller_file" to the generation_dict["real_controller_files"] list
    generation_dict["real_controller_files"] = []
    generation_dict["mock_controller_files"] = []
    for _, controller in generation_dict["controllers"].items():
        generation_dict["real_controller_files"].append(
            controller["real_controller_file"]
        )
        generation_dict["mock_controller_files"].append(
            controller["mock_controller_file"]
        )
        generation_dict["mock_controller_files"].append(controller["mock_signals_file"])

    # add event dispatcher
    real_event_dispatcher_file = os.path.join(
        real_imports_folder_path,
        "controllers",
        "foreign_event_dispatcher.h",
    )
    generation_dict["real_controller_files"].append(real_event_dispatcher_file)
    generation_dict["real_event_dispatcher_file"] = real_event_dispatcher_file

    mock_event_dispatcher_file = os.path.join(
        mock_imports_folder_path,
        "Controllers",
        "EventDispatcher.qml",
    )
    generation_dict["mock_controller_files"].append(mock_event_dispatcher_file)
    generation_dict["mock_event_dispatcher_file"] = mock_event_dispatcher_file

    # add undo redo
    generation_dict["has_undo_redo"] = has_undo_redo
    if has_undo_redo:
        real_undo_redo_controller_file = os.path.join(
            real_imports_folder_path, "controllers", "foreign_undo_redo_controller.h"
        )
        generation_dict["real_controller_files"].append(real_undo_redo_controller_file)
        generation_dict[
            "real_undo_redo_controller_file"
        ] = real_undo_redo_controller_file

        mock_undo_redo_controller_file = os.path.join(
            mock_imports_folder_path, "Controllers", "UndoRedoController.qml"
        )
        generation_dict["mock_controller_files"].append(mock_undo_redo_controller_file)
        generation_dict[
            "mock_undo_redo_controller_file"
        ] = mock_undo_redo_controller_file

        mock_undo_redo_signals_file = os.path.join(
            mock_imports_folder_path, "Controllers", "UndoRedoSignals.qml"
        )
        generation_dict["mock_controller_files"].append(mock_undo_redo_signals_file)
        generation_dict["mock_undo_redo_signals_file"] = mock_undo_redo_signals_file

    # progress signals
    mock_progress_signals_file = os.path.join(
        mock_imports_folder_path, "Controllers", "ProgressSignals.qml"
    )
    generation_dict["mock_controller_files"].append(mock_progress_signals_file)
    generation_dict["mock_progress_signals_file"] = mock_progress_signals_file

    # error signals
    mock_error_signals_file = os.path.join(
        mock_imports_folder_path, "Controllers", "ErrorSignals.qml"
    )
    generation_dict["mock_controller_files"].append(mock_error_signals_file)
    generation_dict["mock_error_signals_file"] = mock_error_signals_file

    # QCoro::QmlTask mock
    qcoro_qmltask_file = os.path.join(
        mock_imports_folder_path, "Controllers", "QCoroQmlTask.qml"
    )
    generation_dict["mock_controller_files"].append(qcoro_qmltask_file)
    generation_dict["qcoro_qmltask_file"] = qcoro_qmltask_file

    # add models
    generation_dict["models"] = []
    generation_dict["mock_model_files"] = []
    generation_dict["real_model_files"] = []

    generation_dict["create_undo_and_redo_singles"] = create_undo_and_redo_singles

    for list_model in list_models:
        list_model_name = list_model["name"]

        entity_name = list_model["entity"]
        entity_name_snake = stringcase.snakecase(entity_name)
        entity_name_pascal = stringcase.pascalcase(entity_name)
        entity_name_spinal = stringcase.spinalcase(entity_name)
        entity_name_camel = stringcase.camelcase(entity_name)

        # displayed_field
        displayed_field = list_model.get("displayed_field", "id")
        displayed_field_snake = stringcase.snakecase(displayed_field)
        displayed_field_pascal = stringcase.pascalcase(displayed_field)
        displayed_field_spinal = stringcase.spinalcase(displayed_field)
        displayed_field_camel = stringcase.camelcase(displayed_field)

        related_name = list_model.get("in_relation_of", "")
        related_name_snake = stringcase.snakecase(related_name)
        related_name_pascal = stringcase.pascalcase(related_name)
        related_name_spinal = stringcase.spinalcase(related_name)
        related_name_camel = stringcase.camelcase(related_name)

        related_field_name = list_model.get("relation_field_name", "")
        related_field_name_snake = stringcase.snakecase(related_field_name)
        related_field_name_pascal = stringcase.pascalcase(related_field_name)
        related_field_name_spinal = stringcase.spinalcase(related_field_name)
        related_field_name_camel = stringcase.camelcase(related_field_name)
        is_related_list = related_name != "" and related_field_name != ""

        list_model_name = list_model["name"]
        if list_model_name == "auto":
            if is_related_list:
                list_model_name = (
                    entity_name_pascal
                    + "ListModelFrom"
                    + related_name_pascal
                    + related_field_name_pascal
                )
            else:
                list_model_name = entity_name_pascal + "ListModel"

        list_model_pascal_name = stringcase.pascalcase(list_model_name)
        list_model_snake_name = stringcase.snakecase(list_model_name)
        list_model_spinal_name = stringcase.spinalcase(list_model_name)
        list_model_camel_name = stringcase.camelcase(list_model_name)

        mock_model_file = os.path.join(
            mock_imports_folder_path,
            "Models",
            f"{list_model_pascal_name}.qml",
        )
        generation_dict["mock_model_files"].append(mock_model_file)
        real_model_file = os.path.join(
            real_imports_folder_path,
            "models",
            f"foreign_{list_model_snake_name}_model.h",
        )
        generation_dict["real_model_files"].append(real_model_file)

        generation_dict["models"].append(
            {
                "mock_model_file": mock_model_file,
                "mock_template_path": "QML/mock_imports/models/",
                "mock_template": "list_model.qml.jinja2",
                "real_model_file": real_model_file,
                "real_template_path": "QML/real_imports/models/",
                "real_template": "foreign_model.h.jinja2",
                "list_model_pascal_name": list_model_pascal_name,
                "list_model_snake_name": list_model_snake_name,
                "list_model_spinal_name": list_model_spinal_name,
                "list_model_camel_name": list_model_camel_name,
                "relation_name_camel": stringcase.camelcase(
                    list_model.get("in_relation_of", "")
                ),
                "is_in_relation_of_another_entity": bool(
                    list_model.get("in_relation_of", False)
                ),
                "fields": tools.get_fields_without_foreign_entities(
                    entities_by_name[list_model["entity"]]["fields"],
                    entities_by_name,
                    list_model["entity"],
                ),
            }
        )

    # add qmldir:
    model_qmldir_file = os.path.join(mock_imports_folder_path, "Models", "qmldir")

    generation_dict["model_qmldir_file"] = model_qmldir_file

    # add CMakelists.txt:
    model_cmakelists_file = os.path.join(
        real_imports_folder_path, "models", "CMakeLists.txt"
    )
    generation_dict["model_cmakelists_file"] = model_cmakelists_file

    # add application_name
    generation_dict["application_name"] = application_name
    generation_dict["application_cpp_domain_name"] = application_cpp_domain_name

    return generation_dict


def _generate_mock_controller_file(
    root_path: str,
    controller: dict,
    generation_dict: dict,
    files_to_be_generated: dict[str, bool],
):
    # generate the mock controller file if in the files_to_be_generated dict the value is True
    if not files_to_be_generated.get(controller["mock_controller_file"], False):
        return

    # Create the jinja2 environment
    template_path = os.path.join("templates", controller["mock_template_path"])
    env = Environment(loader=FileSystemLoader(template_path))
    # Load the template
    template = env.get_template(controller["mock_template"])

    # Render the template
    output = template.render(
        controller=controller,
    )

    output_file = os.path.join(root_path, controller["mock_controller_file"])

    # Create the directory if it does not exist
    os.makedirs(os.path.dirname(output_file), exist_ok=True)

    # Write the output to the file
    with open(output_file, "w") as fh:
        fh.write(output)

    print(f"Successfully wrote file {output_file}")


def _generate_mock_signals_file(
    root_path: str,
    controller: dict,
    generation_dict: dict,
    files_to_be_generated: dict[str, bool],
):
    # generate the mock signals file if in the files_to_be_generated dict the value is True
    if not files_to_be_generated.get(controller["mock_signals_file"], False):
        return

    # Create the jinja2 environment
    template_path = os.path.join("templates", controller["mock_template_path"])
    env = Environment(loader=FileSystemLoader(template_path))
    # Load the template
    template = env.get_template(controller["mock_signals_template"])

    # Render the template
    output = template.render(
        controller=controller,
    )

    output_file = os.path.join(root_path, controller["mock_signals_file"])

    # Create the directory if it does not exist
    os.makedirs(os.path.dirname(output_file), exist_ok=True)

    # Write the output to the file
    with open(output_file, "w") as fh:
        fh.write(output)

    print(f"Successfully wrote file {output_file}")


def _generate_mock_undo_redo_controller_file(
    root_path: str,
    generation_dict: dict,
    files_to_be_generated: dict[str, bool],
):
    # generate the mock undo redo controller file if in the files_to_be_generated dict the value is True
    undo_redo_controller_file = generation_dict["mock_undo_redo_controller_file"]

    if not files_to_be_generated.get(undo_redo_controller_file, False):
        return

    undo_redo_controller_file = os.path.join(
        root_path,
        undo_redo_controller_file,
    )

    # Create the jinja2 environment
    template_path = os.path.join("templates", "QML", "mock_imports", "controllers")
    env = Environment(loader=FileSystemLoader(template_path))
    # Load the template
    template = env.get_template("undo_redo_controller.qml.jinja2")

    # Render the template
    output = template.render()

    # Create the directory if it does not exist
    os.makedirs(os.path.dirname(undo_redo_controller_file), exist_ok=True)

    # Write the output to the file
    with open(undo_redo_controller_file, "w") as fh:
        fh.write(output)

    print(f"Successfully wrote file {undo_redo_controller_file}")


def _generate_mock_undo_redo_signals_file(
    root_path: str,
    generation_dict: dict,
    files_to_be_generated: dict[str, bool],
):
    # generate the mock undo redo signals file if in the files_to_be_generated dict the value is True
    undo_redo_signals_file = generation_dict["mock_undo_redo_signals_file"]

    if not files_to_be_generated.get(undo_redo_signals_file, False):
        return

    undo_redo_signals_file = os.path.join(
        root_path,
        undo_redo_signals_file,
    )

    # Create the jinja2 environment
    template_path = os.path.join("templates", "QML", "mock_imports", "controllers")
    env = Environment(loader=FileSystemLoader(template_path))
    # Load the template
    template = env.get_template("undo_redo_signals.qml.jinja2")

    # Render the template
    output = template.render()

    # Create the directory if it does not exist
    os.makedirs(os.path.dirname(undo_redo_signals_file), exist_ok=True)

    # Write the output to the file
    with open(undo_redo_signals_file, "w") as fh:
        fh.write(output)

    print(f"Successfully wrote file {undo_redo_signals_file}")


def _generate_mock_progress_signals_file(
    root_path: str,
    generation_dict: dict,
    files_to_be_generated: dict[str, bool],
):
    # generate the mock progress signals file if in the files_to_be_generated dict the value is True
    progress_signals_file = generation_dict["mock_progress_signals_file"]

    if not files_to_be_generated.get(progress_signals_file, False):
        return

    progress_signals_file = os.path.join(
        root_path,
        progress_signals_file,
    )

    # Create the jinja2 environment
    template_path = os.path.join("templates", "QML", "mock_imports", "controllers")
    env = Environment(loader=FileSystemLoader(template_path))
    # Load the template
    template = env.get_template("progress_signals.qml.jinja2")

    # Render the template
    output = template.render()

    # Create the directory if it does not exist
    os.makedirs(os.path.dirname(progress_signals_file), exist_ok=True)

    # Write the output to the file
    with open(progress_signals_file, "w") as fh:
        fh.write(output)

    print(f"Successfully wrote file {progress_signals_file}")


def _generate_mock_error_signals_file(
    root_path: str,
    generation_dict: dict,
    files_to_be_generated: dict[str, bool],
):
    # generate the mock error signals file if in the files_to_be_generated dict the value is True
    error_signals_file = generation_dict["mock_error_signals_file"]

    if not files_to_be_generated.get(error_signals_file, False):
        return

    error_signals_file = os.path.join(
        root_path,
        error_signals_file,
    )

    # Create the jinja2 environment
    template_path = os.path.join("templates", "QML", "mock_imports", "controllers")
    env = Environment(loader=FileSystemLoader(template_path))
    # Load the template
    template = env.get_template("error_signals.qml.jinja2")

    # Render the template
    output = template.render()

    # Create the directory if it does not exist
    os.makedirs(os.path.dirname(error_signals_file), exist_ok=True)

    # Write the output to the file
    with open(error_signals_file, "w") as fh:
        fh.write(output)

    print(f"Successfully wrote file {error_signals_file}")


def _generate_mock_event_dispatcher_file(
    root_path: str, generation_dict: dict, files_to_be_generated: dict[str, bool]
):
    # generate the mock event dispatcher file if in the files_to_be_generated dict the value is True
    event_dispatcher_file = generation_dict["mock_event_dispatcher_file"]

    if not files_to_be_generated.get(event_dispatcher_file, False):
        return

    event_dispatcher_file = os.path.join(
        root_path,
        event_dispatcher_file,
    )

    # Create the jinja2 environment
    template_path = os.path.join("templates", "QML", "mock_imports", "controllers")
    env = Environment(loader=FileSystemLoader(template_path))
    # Load the template
    template = env.get_template("event_dispatcher.qml.jinja2")

    # Render the template
    output = template.render(
        controllers=generation_dict["controllers"],
        has_undo_redo=generation_dict["has_undo_redo"],
    )

    # Create the directory if it does not exist
    os.makedirs(os.path.dirname(event_dispatcher_file), exist_ok=True)

    # Write the output to the file
    with open(event_dispatcher_file, "w") as fh:
        fh.write(output)

    print(f"Successfully wrote file {event_dispatcher_file}")


def _generate_mock_qcoro_qmltask_file(
    root_path: str, generation_dict: dict, files_to_be_generated: dict[str, bool]
):
    # generate the mock qcoro qmltask file if in the files_to_be_generated dict the value is True
    qcoro_qmltask_file = generation_dict["qcoro_qmltask_file"]

    if not files_to_be_generated.get(qcoro_qmltask_file, False):
        return

    qcoro_qmltask_file = os.path.join(
        root_path,
        qcoro_qmltask_file,
    )

    # Create the jinja2 environment
    template_path = os.path.join("templates", "QML", "mock_imports", "controllers")
    env = Environment(loader=FileSystemLoader(template_path))
    # Load the template
    template = env.get_template("qcoro_qmltask.qml.jinja2")

    # Render the template
    output = template.render()

    # Create the directory if it does not exist
    os.makedirs(os.path.dirname(qcoro_qmltask_file), exist_ok=True)

    # Write the output to the file
    with open(qcoro_qmltask_file, "w") as fh:
        fh.write(output)

    print(f"Successfully wrote file {qcoro_qmltask_file}")


def _generate_mock_controllers_qmldir_file(
    root_path: str, generation_dict: dict, files_to_be_generated: dict[str, bool]
):
    # generate the mock qmldir file if in the files_to_be_generated dict the value is True
    if not files_to_be_generated.get(generation_dict["controller_qmldir_file"], False):
        return

    # Create the jinja2 environment
    env = Environment(
        loader=FileSystemLoader("templates/QML/mock_imports/controllers/")
    )
    # Load the template
    template = env.get_template("qmldir_template.jinja2")

    singleton_list = []
    for _, controller in generation_dict["controllers"].items():
        name = controller["feature_pascal_name"]
        singleton_list.append(f"singleton {name}Controller 1.0 {name}Controller.qml")
        singleton_list.append(f"singleton {name}Signals 1.0 {name}Signals.qml")

    if generation_dict["has_undo_redo"]:
        singleton_list.append(
            f"singleton UndoRedoController 1.0 UndoRedoController.qml"
        )
        singleton_list.append(f"singleton UndoRedoSignals 1.0 UndoRedoSignals.qml")
    singleton_list.append(f"singleton ProgressSignals 1.0 ProgressSignals.qml")
    singleton_list.append(f"singleton ErrorSignals 1.0 ErrorSignals.qml")
    singleton_list.append(f"singleton EventDispatcher 1.0 EventDispatcher.qml")
    singleton_list.append(f"QCoroQmlTask 1.0 QCoroQmlTask.qml")

    # Render the template
    output = template.render(singleton_list=singleton_list)

    output_file = os.path.join(root_path, generation_dict["controller_qmldir_file"])

    # Create the directory if it does not exist
    os.makedirs(os.path.dirname(output_file), exist_ok=True)

    # Write the output to the file
    with open(output_file, "w") as fh:
        fh.write(output)

    print(f"Successfully wrote file {output_file}")


def _generate_real_controller_file(
    root_path: str,
    controller: dict,
    generation_dict: dict,
    files_to_be_generated: dict[str, bool],
    uncrustify_config_file: str,
):
    # generate the real controller file if in the files_to_be_generated dict the value is True
    real_controller_file = controller["real_controller_file"]

    if not files_to_be_generated.get(real_controller_file, False):
        return

    # Create the jinja2 environment
    template_path = os.path.join("templates", controller["real_template_path"])
    env = Environment(loader=FileSystemLoader(template_path))
    # Load the template
    template = env.get_template(controller["real_template"])

    # Render the template
    output = template.render(
        controller=controller,
        application_cpp_domain_name=generation_dict["application_cpp_domain_name"],
    )

    output_file = os.path.join(root_path, real_controller_file)

    # Create the directory if it does not exist
    os.makedirs(os.path.dirname(output_file), exist_ok=True)

    # Write the output to the file
    with open(output_file, "w") as fh:
        fh.write(output)

    # if uncrustify_config_file:
    #     uncrustify.run_uncrustify(real_controller_file, uncrustify_config_file)
    clang_format.run_clang_format(output_file)

    print(f"Successfully wrote file {output_file}")


def _generate_real_undo_redo_controller_file(
    root_path: str, generation_dict: dict, files_to_be_generated: dict[str, bool]
):
    # generate the real undo redo controller file if in the files_to_be_generated dict the value is True
    undo_redo_controller_file = generation_dict["real_undo_redo_controller_file"]

    if not files_to_be_generated.get(undo_redo_controller_file, False):
        return

    undo_redo_controller_file = os.path.join(
        root_path,
        undo_redo_controller_file,
    )

    # Create the jinja2 environment
    template_path = os.path.join("templates", "QML", "real_imports", "controllers")
    env = Environment(loader=FileSystemLoader(template_path))
    # Load the template
    template = env.get_template("foreign_undo_redo_controller.h.jinja2")

    # Render the template
    output = template.render(
        application_cpp_domain_name=generation_dict["application_cpp_domain_name"],
    )

    # Create the directory if it does not exist
    os.makedirs(os.path.dirname(undo_redo_controller_file), exist_ok=True)

    # Write the output to the file
    with open(undo_redo_controller_file, "w") as fh:
        fh.write(output)

    # if uncrustify_config_file:
    #     uncrustify.run_uncrustify(undo_redo_controller_file, uncrustify_config_file)
    clang_format.run_clang_format(undo_redo_controller_file)

    print(f"Successfully wrote file {undo_redo_controller_file}")


def _generate_real_event_dispatcher_file(
    root_path: str, generation_dict: dict, files_to_be_generated: dict[str, bool]
):
    # generate the real event dispatcher file if in the files_to_be_generated dict the value is True
    event_dispatcher_file = generation_dict["real_event_dispatcher_file"]

    if not files_to_be_generated.get(event_dispatcher_file, False):
        return

    event_dispatcher_file = os.path.join(
        root_path,
        event_dispatcher_file,
    )

    # Create the jinja2 environment
    template_path = os.path.join("templates", "QML", "real_imports", "controllers")
    env = Environment(loader=FileSystemLoader(template_path))
    # Load the template
    template = env.get_template("foreign_event_dispatcher.h.jinja2")

    # Render the template
    output = template.render(
        application_cpp_domain_name=generation_dict["application_cpp_domain_name"],
    )

    # Create the directory if it does not exist
    os.makedirs(os.path.dirname(event_dispatcher_file), exist_ok=True)

    # Write the output to the file
    with open(event_dispatcher_file, "w") as fh:
        fh.write(output)

    # if uncrustify_config_file:
    #     uncrustify.run_uncrustify(event_dispatcher_file, uncrustify_config_file)
    clang_format.run_clang_format(event_dispatcher_file)

    print(f"Successfully wrote file {event_dispatcher_file}")


def _generate_real_controllers_cmakelists_file(
    root_path: str, generation_dict: dict, files_to_be_generated: dict[str, bool]
):
    controller_cmakelists_file = generation_dict["controller_cmakelists_file"]

    # generate the real cmakelists file if in the files_to_be_generated dict the value is True
    if not files_to_be_generated.get(controller_cmakelists_file, False):
        return

    output_file = os.path.join(root_path, controller_cmakelists_file)

    # Create the jinja2 environment
    env = Environment(
        loader=FileSystemLoader("templates/QML/real_imports/controllers/")
    )
    # Load the template
    template = env.get_template("cmakelists.txt.jinja2")

    files = generation_dict["real_controller_files"]
    relative_files = []
    for file in files:
        relative_files.append(
            os.path.relpath(
                os.path.join(root_path, file), os.path.dirname(output_file)
            ).replace("\\", "/")
        )

    # Render the template
    output = template.render(
        files=relative_files,
        application_name=generation_dict["application_name"],
    )

    # Create the directory if it does not exist
    os.makedirs(os.path.dirname(output_file), exist_ok=True)

    # Write the output to the file
    with open(output_file, "w") as fh:
        fh.write(output)

    print(f"Successfully wrote file {output_file}")


def _generate_mock_model_file(
    root_path: str,
    model: dict,
    generation_dict: dict,
    files_to_be_generated: dict[str, bool],
):
    # generate the mock model file if in the files_to_be_generated dict the value is True
    if not files_to_be_generated.get(model["mock_model_file"], False):
        return

    # Create the jinja2 environment
    template_path = os.path.join("templates", model["mock_template_path"])
    env = Environment(loader=FileSystemLoader(template_path))
    # Load the template
    template = env.get_template(model["mock_template"])

    # Render the template
    output = template.render(
        model=model,
    )

    output_file = os.path.join(root_path, model["mock_model_file"])

    # Create the directory if it does not exist
    os.makedirs(os.path.dirname(output_file), exist_ok=True)

    # Write the output to the file
    with open(output_file, "w") as fh:
        fh.write(output)

    print(f"Successfully wrote file {output_file}")


def _generate_real_model_file(
    root_path: str,
    model: dict,
    generation_dict: dict,
    files_to_be_generated: dict[str, bool],
):
    # generate the real model file if in the files_to_be_generated dict the value is True
    real_model_file = model["real_model_file"]

    if not files_to_be_generated.get(real_model_file, False):
        return

    # Create the jinja2 environment
    template_path = os.path.join("templates", model["real_template_path"])
    env = Environment(loader=FileSystemLoader(template_path))
    # Load the template
    template = env.get_template(model["real_template"])

    # Render the template
    output = template.render(
        model=model,
        application_cpp_domain_name=generation_dict["application_cpp_domain_name"],
    )

    output_file = os.path.join(root_path, real_model_file)

    # Create the directory if it does not exist
    os.makedirs(os.path.dirname(output_file), exist_ok=True)

    # Write the output to the file
    with open(output_file, "w") as fh:
        fh.write(output)

    # if uncrustify_config_file:
    #     uncrustify.run_uncrustify(real_model_file, uncrustify_config_file)
    clang_format.run_clang_format(output_file)

    print(f"Successfully wrote file {output_file}")


def _generate_mock_models_qmldir_file(
    root_path: str, generation_dict: dict, files_to_be_generated: dict[str, bool]
):
    # generate the mock qmldir file if in the files_to_be_generated dict the value is True
    if not files_to_be_generated.get(generation_dict["model_qmldir_file"], False):
        return

    # Create the jinja2 environment
    env = Environment(loader=FileSystemLoader("templates/QML/mock_imports/models/"))
    # Load the template
    template = env.get_template("qmldir_template.jinja2")

    models_to_declare_list = []
    for model in generation_dict["models"]:
        name = model["list_model_pascal_name"]
        models_to_declare_list.append(f"{name} 1.0 {name}.qml")

    # Render the template
    output = template.render(models_to_declare_list=models_to_declare_list)

    output_file = os.path.join(root_path, generation_dict["model_qmldir_file"])

    # Create the directory if it does not exist
    os.makedirs(os.path.dirname(output_file), exist_ok=True)

    # Write the output to the file
    with open(output_file, "w") as fh:
        fh.write(output)

    print(f"Successfully wrote file {output_file}")


def _generate_real_models_cmakelists_file(
    root_path: str, generation_dict: dict, files_to_be_generated: dict[str, bool]
):
    controller_cmakelists_file = generation_dict["model_cmakelists_file"]

    # generate the real cmakelists file if in the files_to_be_generated dict the value is True
    if not files_to_be_generated.get(controller_cmakelists_file, False):
        return

    output_file = os.path.join(root_path, controller_cmakelists_file)

    # Create the jinja2 environment
    env = Environment(loader=FileSystemLoader("templates/QML/real_imports/models/"))
    # Load the template
    template = env.get_template("cmakelists.txt.jinja2")

    files = generation_dict["real_model_files"]
    relative_files = []
    for file in files:
        relative_files.append(
            os.path.relpath(
                os.path.join(root_path, file), os.path.dirname(output_file)
            ).replace("\\", "/")
        )

    # Render the template
    output = template.render(
        files=relative_files,
        application_name=generation_dict["application_name"],
    )

    # Create the directory if it does not exist
    os.makedirs(os.path.dirname(output_file), exist_ok=True)

    # Write the output to the file
    with open(output_file, "w") as fh:
        fh.write(output)

    print(f"Successfully wrote file {output_file}")


def generate_qml_files(
    root_path: str,
    manifest_file: str,
    files_to_be_generated: dict[str, bool] = {},
    uncrustify_config_file: str = None,
):
    with open(manifest_file, "r") as stream:
        try:
            manifest_data = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)
            return

    application_name = manifest_data.get("global", {}).get(
        "application_name", "example"
    )
    application_name = stringcase.spinalcase(application_name)

    application_cpp_domain_name = manifest_data.get("global", {}).get(
        "application_cpp_domain_name", "Example"
    )

    entities_data = manifest_data.get("entities", {})
    entities_list = entities_data.get("list", [])
    # remove entities that are not to be generated
    entities_list = [entity for entity in entities_list]

    # Organize entities by name for easier lookup
    entities_by_name = {entity["name"]: entity for entity in entities_list}

    application_data = manifest_data.get("application", {})
    feature_list = application_data.get("features", [])

    # Organize feature_list by name for easier lookup
    feature_by_name = {feature["name"]: feature for feature in feature_list}

    controller_data = manifest_data.get("controller", {})
    has_undo_redo = controller_data.get("create_undo_redo_controller", False)

    presenter_data = manifest_data.get("presenter", {})
    create_undo_and_redo_singles = presenter_data.get(
        "create_undo_and_redo_singles", False
    )
    list_models = presenter_data.get("list_models", [])
    singles = presenter_data.get("singles", [])

    qml_data = manifest_data.get("qml", [])

    generation_dict = _get_generation_dict(
        qml_data["real_imports_folder_path"],
        qml_data["mock_imports_folder_path"],
        feature_by_name,
        entities_by_name,
        has_undo_redo,
        create_undo_and_redo_singles,
        list_models,
        singles,
        application_name,
        application_cpp_domain_name,
    )

    # generate mock files
    for _, controller in generation_dict["controllers"].items():
        _generate_mock_controller_file(
            root_path, controller, generation_dict, files_to_be_generated
        )
        _generate_mock_signals_file(
            root_path, controller, generation_dict, files_to_be_generated
        )

    _generate_mock_event_dispatcher_file(
        root_path, generation_dict, files_to_be_generated
    )
    if has_undo_redo:
        _generate_mock_undo_redo_controller_file(
            root_path, generation_dict, files_to_be_generated
        )
        _generate_mock_undo_redo_signals_file(
            root_path, generation_dict, files_to_be_generated
        )

    _generate_mock_progress_signals_file(
        root_path, generation_dict, files_to_be_generated
    )
    _generate_mock_error_signals_file(root_path, generation_dict, files_to_be_generated)

    _generate_mock_qcoro_qmltask_file(root_path, generation_dict, files_to_be_generated)

    # generate mock qmldir file
    _generate_mock_controllers_qmldir_file(
        root_path, generation_dict, files_to_be_generated
    )

    # generate real files
    for _, controller in generation_dict["controllers"].items():
        _generate_real_controller_file(
            root_path,
            controller,
            generation_dict,
            files_to_be_generated,
            uncrustify_config_file,
        )

    _generate_real_event_dispatcher_file(
        root_path, generation_dict, files_to_be_generated
    )

    if has_undo_redo:
        _generate_real_undo_redo_controller_file(
            root_path, generation_dict, files_to_be_generated
        )

    # generate real CMakeLists.txt file
    _generate_real_controllers_cmakelists_file(
        root_path, generation_dict, files_to_be_generated
    )

    # models
    for model in generation_dict["models"]:
        _generate_mock_model_file(
            root_path, model, generation_dict, files_to_be_generated
        )
        _generate_real_model_file(
            root_path, model, generation_dict, files_to_be_generated
        )

    # generate mock qmldir file
    _generate_mock_models_qmldir_file(root_path, generation_dict, files_to_be_generated)
    _generate_real_models_cmakelists_file(
        root_path, generation_dict, files_to_be_generated
    )


def get_files_to_be_generated(
    manifest_file: str, files_to_be_generated: dict[str, bool] = {}
) -> list[str]:
    """
    Get the list of files that need to be generated based on the manifest file
    """
    # Read the manifest file
    with open(manifest_file, "r") as stream:
        try:
            manifest_data = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)
            return

    application_name = manifest_data.get("global", {}).get(
        "application_name", "example"
    )
    application_name = stringcase.spinalcase(application_name)

    application_cpp_domain_name = manifest_data.get("global", {}).get(
        "application_cpp_domain_name", "Example"
    )

    entities_data = manifest_data.get("entities", {})
    entities_list = entities_data.get("list", [])
    # remove entities that are not to be generated
    entities_list = [entity for entity in entities_list]

    # Organize entities by name for easier lookup
    entities_by_name = {entity["name"]: entity for entity in entities_list}

    application_data = manifest_data.get("application", {})

    controller_data = manifest_data.get("controller", {})
    has_undo_redo = controller_data.get("create_undo_redo_controller", False)

    feature_list = application_data.get("features", [])

    presenter_data = manifest_data.get("presenter", {})
    create_undo_and_redo_singles = presenter_data.get(
        "create_undo_and_redo_singles", False
    )

    list_models = presenter_data.get("list_models", [])
    singles = presenter_data.get("singles", [])

    # Organize feature_list by name for easier lookup
    feature_by_name = {feature["name"]: feature for feature in feature_list}

    qml_data = manifest_data.get("qml", [])

    files = []
    generation_dict = _get_generation_dict(
        qml_data["real_imports_folder_path"],
        qml_data["mock_imports_folder_path"],
        feature_by_name,
        entities_by_name,
        has_undo_redo,
        create_undo_and_redo_singles,
        list_models,
        singles,
        application_name,
        application_cpp_domain_name,
    )
    files += generation_dict["real_controller_files"]
    files += generation_dict["mock_controller_files"]

    # for _, controller in generation_dict["controllers"].items():
    #     files += feature["real_controller_files"]

    # # add CMakelists.txt:
    controller_cmakelists_file = generation_dict["controller_cmakelists_file"]
    files.append(controller_cmakelists_file)

    # # add qmldir:
    controller_qmldir_file = generation_dict["controller_qmldir_file"]
    files.append(controller_qmldir_file)

    # list models 
    files += generation_dict["real_model_files"]
    files += generation_dict["mock_model_files"]

    # # add CMakelists.txt:
    model_cmakelists_file = generation_dict["model_cmakelists_file"]
    files.append(model_cmakelists_file)

    # # add qmldir:
    model_qmldir_file = generation_dict["model_qmldir_file"]
    files.append(model_qmldir_file)

    # strip from files if the value in files_to_be_generated is False
    if files_to_be_generated:
        for path, generate in files_to_be_generated.items():
            if not generate:
                files.remove(path)

    return files


# generate the files into the preview folder
def preview_qml_files(
    root_path: str,
    manifest_file: str,
    files_to_be_generated: dict[str, bool] = {},
    uncrustify_config_file: str = None,
):
    manifest_preview_file = "temp/manifest_preview.yaml"

    # make a copy of the manifest file into temp/manifest_preview.yaml
    shutil.copy(manifest_file, manifest_preview_file)

    # modify the manifest file to generate the files into the preview folder
    with open(manifest_preview_file, "r") as fh:
        manifest = yaml.safe_load(fh)

    # remove .. from the path
    manifest["qml"]["mock_imports_folder_path"] = manifest["qml"][
        "mock_imports_folder_path"
    ].replace("..", "")
    manifest["qml"]["real_imports_folder_path"] = manifest["qml"][
        "real_imports_folder_path"
    ].replace("..", "")

    # write the modified manifest file
    with open(manifest_preview_file, "w") as fh:
        yaml.dump(manifest, fh)

    root_path = os.path.join(root_path, "qleany_preview")

    # remove .. from the path
    if files_to_be_generated:
        preview_files_to_be_generated = {}
        for path, value in files_to_be_generated.items():
            preview_files_to_be_generated[path.replace("..", "")] = value

        generate_qml_files(
            root_path,
            manifest_preview_file,
            preview_files_to_be_generated,
            uncrustify_config_file,
        )

    else:
        generate_qml_files(root_path, manifest_preview_file, {}, uncrustify_config_file)


# Main execution
if __name__ == "__main__":
    full_path = Path(__file__).resolve().parent

    # Add the current directory to the path so that we can import the generated files
    sys.path.append(full_path)

    # Set the current directory to the generator directory
    os.chdir(full_path)

    if len(sys.argv) > 1:
        manifest_arg = sys.argv[1]
        if manifest_arg.endswith(".yaml") or manifest_arg.endswith(".yml"):
            manifest_file = manifest_arg
            root_path = Path(manifest_file).parent

            if len(sys.argv) > 2 and sys.argv[2] == "--preview":
                preview_qml_files(root_path, manifest_file)
            else:
                generate_qml_files(root_path, manifest_file)
        else:
            print("Error: Manifest file must be named 'qleany.yaml' or 'qleany.yml'")
    else:
        print("Error: Please provide the manifest file as an argument")
